from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pyarrow as pa
import pyarrow_hotfix  # noqa: F401

import xorq.vendor.ibis.expr.datatypes as dt
from xorq.vendor.ibis.expr.schema import Schema
from xorq.vendor.ibis.formats import DataMapper, SchemaMapper, TableProxy, TypeMapper


if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl


# _from_pyarrow_types[pa.string_view()] = dt.String
# _from_pyarrow_types[pa.large_string()] = LargeString
# _to_pyarrow_types[LargeString] = pa.large_string()


_from_pyarrow_types = {
    pa.int8(): dt.Int8,
    pa.int16(): dt.Int16,
    pa.int32(): dt.Int32,
    pa.int64(): dt.Int64,
    pa.uint8(): dt.UInt8,
    pa.uint16(): dt.UInt16,
    pa.uint32(): dt.UInt32,
    pa.uint64(): dt.UInt64,
    pa.float16(): dt.Float16,
    pa.float32(): dt.Float32,
    pa.float64(): dt.Float64,
    pa.string(): dt.String,
    pa.binary(): dt.Binary,
    pa.bool_(): dt.Boolean,
    pa.date32(): dt.Date,
    pa.date64(): dt.Date,
    pa.null(): dt.Null,
    pa.large_binary(): dt.Binary,
    pa.large_string(): dt.LargeString,
    pa.binary_view(): dt.BinaryView,
    pa.string_view(): dt.String,
}


_to_pyarrow_types = {
    dt.Null: pa.null(),
    dt.Boolean: pa.bool_(),
    dt.Binary: pa.binary(),
    dt.Int8: pa.int8(),
    dt.Int16: pa.int16(),
    dt.Int32: pa.int32(),
    dt.Int64: pa.int64(),
    dt.UInt8: pa.uint8(),
    dt.UInt16: pa.uint16(),
    dt.UInt32: pa.uint32(),
    dt.UInt64: pa.uint64(),
    dt.Float16: pa.float16(),
    dt.Float32: pa.float32(),
    dt.Float64: pa.float64(),
    dt.String: pa.string(),
    # assume unknown types can be converted into strings
    dt.Unknown: pa.string(),
    dt.MACADDR: pa.string(),
    dt.INET: pa.string(),
    dt.UUID: pa.string(),
    dt.JSON: pa.string(),
    dt.LargeString: pa.large_string(),
    dt.BinaryView: pa.binary_view(),
}


class PyArrowType(TypeMapper):
    @classmethod
    def to_ibis(cls, typ: pa.DataType, nullable=True) -> dt.DataType:
        """Convert a pyarrow type to an ibis type."""
        if pa.types.is_null(typ):
            return dt.null
        elif pa.types.is_decimal(typ):
            return dt.Decimal(typ.precision, typ.scale, nullable=nullable)
        elif pa.types.is_timestamp(typ):
            return dt.Timestamp.from_unit(typ.unit, timezone=typ.tz, nullable=nullable)
        elif pa.types.is_time(typ):
            return dt.Time(nullable=nullable)
        elif pa.types.is_duration(typ):
            return dt.Interval(typ.unit, nullable=nullable)
        elif pa.types.is_interval(typ):
            raise ValueError("Arrow interval type is not supported")
        elif (
            pa.types.is_list(typ)
            or pa.types.is_large_list(typ)
            or pa.types.is_fixed_size_list(typ)
        ):
            value_dtype = cls.to_ibis(typ.value_type, typ.value_field.nullable)
            return dt.Array(value_dtype, nullable=nullable)
        elif pa.types.is_struct(typ):
            field_dtypes = {
                field.name: cls.to_ibis(field.type, field.nullable) for field in typ
            }
            return dt.Struct(field_dtypes, nullable=nullable)
        elif pa.types.is_map(typ):
            # TODO(kszucs): keys_sorted has just been exposed in pyarrow
            key_dtype = cls.to_ibis(typ.key_type, typ.key_field.nullable)
            value_dtype = cls.to_ibis(typ.item_type, typ.item_field.nullable)
            return dt.Map(key_dtype, value_dtype, nullable=nullable)
        elif pa.types.is_dictionary(typ):
            return cls.to_ibis(typ.value_type)
        elif (
            isinstance(typ, pa.ExtensionType)
            and type(typ).__module__ == "geoarrow.types.type_pyarrow"
        ):
            from geoarrow import types as gat

            gat.type_pyarrow.register_extension_types()

            auth_code = None
            if typ.crs is not None:
                crs_dict = typ.crs.to_json_dict()
                if "id" in crs_dict:
                    crs_id = crs_dict["id"]
                    if "authority" in crs_id and "code" in crs_id:
                        auth_code = (crs_id["authority"], crs_id["code"])

            if typ.crs is not None and auth_code is None:
                # It is possible to have PROJJSON that does not have an authority/code
                # attached, either because the producer didn't have that information
                # (e.g., because they were reading a older shapefile). In this case,
                # pyproj can often guess the authority/code.
                import pyproj

                auth_code = pyproj.CRS(typ.crs.to_json()).to_authority()
                if auth_code is None:
                    raise ValueError(f"Can't resolve SRID of crs {typ.crs}")

            if auth_code is None:
                srid = None
            elif auth_code == ("OGC", "CRS84"):
                # OGC:CRS84 and EPSG:4326 are identical except for the order of
                # coordinates (i.e., lon lat vs. lat lon) in their official definition.
                # This axis ordering is ignored in all but the most obscure scenarios
                # such that these are identical. OGC:CRS84 is more correct, but EPSG:4326
                # is more common.
                srid = 4326
            else:
                # This works because the two most common srid authorities are EPSG and ESRI
                # and the "codes" are all integers and don't intersect with each other on
                # purpose. This won't scale to something like OGC:CRS27 (not common).
                srid = int(auth_code[1])

            if typ.edge_type == gat.EdgeType.SPHERICAL:
                geotype = "geography"
            else:
                geotype = "geometry"

            return dt.GeoSpatial(geotype, srid, nullable)
        else:
            return _from_pyarrow_types[typ](nullable=nullable)

    @classmethod
    def from_ibis(cls, dtype: dt.DataType) -> pa.DataType:
        """Convert an ibis type to a pyarrow type."""
        if dtype.is_decimal():
            # set default precision and scale to something; unclear how to choose this
            precision = 38 if dtype.precision is None else dtype.precision
            scale = 9 if dtype.scale is None else dtype.scale

            if precision > 76:
                raise ValueError(
                    f"Unsupported precision {dtype.precision} for decimal type"
                )
            elif precision > 38:
                return pa.decimal256(precision, scale)
            else:
                return pa.decimal128(precision, scale)
        elif dtype.is_timestamp():
            return pa.timestamp(
                dtype.unit.short if dtype.scale is not None else "us", tz=dtype.timezone
            )
        elif dtype.is_interval():
            short = dtype.unit.short
            if short in {"ns", "us", "ms", "s"}:
                return pa.duration(short)
            else:
                return pa.month_day_nano_interval()
        elif dtype.is_time():
            return pa.time64("ns")
        elif dtype.is_date():
            return pa.date32()
        elif dtype.is_array():
            value_field = pa.field(
                "item",
                cls.from_ibis(dtype.value_type),
                nullable=dtype.value_type.nullable,
            )
            return pa.list_(value_field)
        elif dtype.is_struct():
            fields = [
                pa.field(name, cls.from_ibis(dtype), nullable=dtype.nullable)
                for name, dtype in dtype.items()
            ]
            return pa.struct(fields)
        elif dtype.is_map():
            key_field = pa.field(
                "key",
                cls.from_ibis(dtype.key_type),
                nullable=False,  # pyarrow doesn't allow nullable keys
            )
            value_field = pa.field(
                "value",
                cls.from_ibis(dtype.value_type),
                nullable=dtype.value_type.nullable,
            )
            return pa.map_(key_field, value_field, keys_sorted=False)
        elif dtype.is_geospatial():
            from geoarrow import types as gat

            # Resolve CRS
            if dtype.srid is None:
                crs = None
            elif dtype.srid == 4326:
                crs = gat.OGC_CRS84
            else:
                import pyproj

                # Assume that these are EPSG codes. An srid is more accurately a key
                # into a backend/connection-specific lookup table; however, most usage
                # should work with this assumption.
                crs = pyproj.CRS(f"EPSG:{dtype.srid}")

            # Resolve edge type
            if dtype.geotype == "geography":
                edge_type = gat.EdgeType.SPHERICAL
            else:
                edge_type = gat.EdgeType.PLANAR

            return gat.wkb(crs=crs, edge_type=edge_type).to_pyarrow()
        else:
            try:
                return _to_pyarrow_types[type(dtype)]
            except KeyError:
                raise NotImplementedError(
                    f"Converting {dtype} to pyarrow is not supported yet"
                )


class PyArrowSchema(SchemaMapper):
    @classmethod
    def from_ibis(cls, schema: Schema) -> pa.Schema:
        """Convert a schema to a pyarrow schema."""
        fields = [
            pa.field(name, PyArrowType.from_ibis(dtype), nullable=dtype.nullable)
            for name, dtype in schema.items()
        ]
        return pa.schema(fields)

    @classmethod
    def to_ibis(cls, schema: pa.Schema) -> Schema:
        """Convert a pyarrow schema to a schema."""
        fields = [(f.name, PyArrowType.to_ibis(f.type, f.nullable)) for f in schema]
        return Schema.from_tuples(fields)


class PyArrowData(DataMapper):
    @classmethod
    def infer_scalar(cls, scalar: Any) -> dt.DataType:
        """Infer the ibis type of a scalar."""
        return PyArrowType.to_ibis(pa.scalar(scalar).type)

    @classmethod
    def infer_column(cls, column: Sequence) -> dt.DataType:
        """Infer the ibis type of a sequence."""
        if isinstance(column, pa.Array):
            return PyArrowType.to_ibis(column.type)

        try:
            pyarrow_type = pa.array(column, from_pandas=True).type
            # pyarrow_type = pa.infer_type(column, from_pandas=True)
        except pa.ArrowInvalid:
            try:
                # handle embedded series objects
                return dt.highest_precedence(map(dt.infer, column))
            except TypeError:
                # we can still have a type error, e.g., float64 and string in the
                # same array
                return dt.unknown
        except pa.ArrowTypeError:
            # arrow can't infer the type
            return dt.unknown
        else:
            # arrow inferred the type, now convert that type to an ibis type
            return PyArrowType.to_ibis(pyarrow_type)

    @classmethod
    def infer_table(cls, table) -> Schema:
        """Infer the schema of a table."""
        if not isinstance(table, pa.Table):
            table = pa.table(table)

        return PyArrowSchema.to_ibis(table.schema)

    @classmethod
    def convert_scalar(cls, scalar: pa.Scalar, dtype: dt.DataType) -> pa.Scalar:
        desired_type = PyArrowType.from_ibis(dtype)
        scalar_type = scalar.type
        if scalar_type != desired_type:
            try:
                return scalar.cast(desired_type)
            except pa.ArrowNotImplementedError:
                # pyarrow doesn't support some scalar casts that are supported
                # when using arrays or tables
                return pa.array([scalar.as_py()], type=scalar_type).cast(desired_type)[
                    0
                ]
        else:
            return scalar

    @classmethod
    def convert_column(cls, column: pa.Array, dtype: dt.DataType) -> pa.Array:
        desired_type = PyArrowType.from_ibis(dtype)
        if column.type != desired_type:
            return column.cast(desired_type)
        else:
            return column

    @classmethod
    def convert_table(cls, table: pa.Table, schema: Schema) -> pa.Table:
        desired_schema = PyArrowSchema.from_ibis(schema)
        pa_schema = table.schema

        if pa_schema != desired_schema:
            return table.cast(desired_schema, safe=False)
        else:
            return table


class PyArrowTableProxy(TableProxy):
    def proxy_equals(self, other: "PyArrowTableProxy") -> bool:
        return self.obj.equals(other.obj)

    def to_frame(self):
        return self.obj.to_pandas()

    def to_pyarrow(self, schema: Schema) -> pa.Table:
        return self.obj

    def to_polars(self, schema: Schema) -> pl.DataFrame:
        import polars as pl

        from xorq.vendor.ibis.formats.polars import PolarsData

        df = pl.from_arrow(self.obj)
        return PolarsData.convert_table(df, schema)



from __future__ import annotations

import contextlib
import sys
import tempfile
from collections.abc import Callable
from html import escape
from typing import Optional

import graphviz as gv

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor import ibis
from xorq.vendor.ibis.common.graph import Graph


def get_type(node):
    with contextlib.suppress(AttributeError, NotImplementedError):
        return escape(str(node.dtype))

    try:
        schema = node.schema
    except (AttributeError, NotImplementedError):
        # TODO(kszucs): this branch should be removed
        try:
            # As a last resort try get the name of the output_type class
            return node.output_type.__name__
        except (AttributeError, NotImplementedError):
            return "\u2205"  # empty set character
    except com.XorqError:
        assert isinstance(node, ops.Join)
        left_table_name = getattr(node.left, "name", None) or ops.genname()
        left_schema = node.left.schema
        right_table_name = getattr(node.right, "name", None) or ops.genname()
        right_schema = node.right.schema
        pairs = [
            (f"{left_table_name}.{left_column}", type)
            for left_column, type in left_schema.items()
        ] + [
            (f"{right_table_name}.{right_column}", type)
            for right_column, type in right_schema.items()
        ]
        schema = ibis.schema(pairs)
    else:
        # Simple relations have the same schema as their parent so avoid
        # re-rendering the same schema fields for these relations
        if isinstance(node, ops.relations.Simple):
            return '<BR ALIGN="LEFT" />:: …'

    return '<BR ALIGN="LEFT" />' + '<BR ALIGN="LEFT" />'.join(
        f"<I>{escape(name)}</I>: {escape(str(type))}"
        for name, type in zip(schema.names, schema.types)
    )


def get_label(node):
    typename = get_type(node)  # Already an escaped string
    name = type(node).__name__
    nodename = (
        node.name
        if isinstance(
            node,
            (
                ops.Literal,
                ops.Field,
                ops.Alias,
                ops.PhysicalTable,
            ),
        )
        else None
    )
    if nodename is not None:
        if isinstance(node, ops.Relation):
            label_fmt = "<<I>{}</I>: <B>{}</B>{}>"
        else:
            label_fmt = '<<I>{}</I>: <B>{}</B><BR ALIGN="LEFT" />:: {}>'
        # typename is already escaped
        label = label_fmt.format(escape(nodename), escape(name), typename)
    else:
        if isinstance(node, ops.Relation):
            label_fmt = "<<B>{}</B>{}>"
        else:
            label_fmt = '<<B>{}</B><BR ALIGN="LEFT" />:: {}>'
        label = label_fmt.format(escape(name), typename)
    return label


DEFAULT_NODE_ATTRS = {"shape": "box", "fontname": "Deja Vu Sans Mono"}
DEFAULT_EDGE_ATTRS = {"fontname": "Deja Vu Sans Mono"}

NodeAttributeGetter = Callable[[ops.Node], Optional[dict[str, str]]]
EdgeAttributeGetter = Callable[[ops.Node, ops.Node], Optional[dict[str, str]]]


def to_graph(
    expr,
    node_attr=None,
    node_attr_getter: NodeAttributeGetter | None = None,
    edge_attr=None,
    edge_attr_getter: EdgeAttributeGetter | None = None,
    label_edges: bool = False,
):
    graph = Graph.from_bfs(expr.op(), filter=ops.Node)

    g = gv.Digraph(
        node_attr=DEFAULT_NODE_ATTRS | (node_attr or {}),
        edge_attr=DEFAULT_EDGE_ATTRS | (edge_attr or {}),
    )

    g.attr(rankdir="BT")

    seen = set()
    edges = set()

    for v, us in graph.items():
        vhash = str(hash(v))
        if v not in seen:
            g.node(
                vhash,
                label=get_label(v),
                _attributes=node_attr_getter(v) if node_attr_getter else {},
            )
            seen.add(v)

        for u in us:
            uhash = str(hash(u))
            if u not in seen:
                g.node(
                    uhash,
                    label=get_label(u),
                    _attributes=node_attr_getter(u) if node_attr_getter else {},
                )
                seen.add(u)
            if (edge := (u, v)) not in edges:
                if not label_edges:
                    label = None
                else:
                    if isinstance(v, ops.Relation):
                        if (name := getattr(u, "name", None)) in v.fields:
                            name = f"fields[{name!r}]"
                        else:
                            name = None
                    else:
                        for name, arg in zip(v.argnames, v.args):
                            if isinstance(arg, tuple) and u in arg:
                                index = arg.index(u)
                                name = f"{name}[{index}]"
                                break
                            elif arg == u:
                                break
                        else:
                            name = None

                    if name is not None:
                        label = f"<.{name}>"
                    else:
                        label = None

                g.edge(
                    uhash,
                    vhash,
                    label=label,
                    _attributes=edge_attr_getter(u, v) if edge_attr_getter else {},
                )
                edges.add(edge)
    return g


def draw(graph, path=None, format="png", verbose: bool = False):
    if verbose:
        print(graph.source, file=sys.stderr)  # noqa: T201

    piped_source = graph.pipe(format=format)

    if path is None:
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=f".{format}", mode="wb"
        ) as f:
            f.write(piped_source)
        return f.name
    else:
        with open(path, mode="wb") as f:
            f.write(piped_source)
        return path


if __name__ == "__main__":
    import json
    from argparse import ArgumentParser

    from xorq.vendor.ibis import _

    p = ArgumentParser(
        description="Render a GraphViz SVG of an example ibis expression."
    )

    p.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Print GraphViz DOT code to stderr.",
    )
    p.add_argument(
        "-l",
        "--label-edges",
        action="store_true",
        help="Show operation inputs as edge labels.",
    )
    p.add_argument(
        "-n",
        "--node-attr",
        type=lambda x: json.loads(x) if x else {},
        default="{}",
        help='JSON string of node attributes. E.g., \'{"fontname": "Roboto Mono", "fontsize": "10"}\'',
    )
    p.add_argument(
        "-e",
        "--edge-attr",
        type=lambda x: json.loads(x) if x else {},
        default="{}",
        help='JSON string of edge attributes. E.g., \'{"fontsize": "8"}\'',
    )

    args = p.parse_args()

    left = ibis.table({"a": "int64", "b": "string"}, name="left")
    right = ibis.table({"b": "string", "c": "int64", "d": "string"}, name="right")
    expr = (
        left.inner_join(right, "b")
        .select(left.a, b=right.c, c=right.d)
        .filter((_.a + _.b * 2 * _.b / _.b**3 > 4) & (_.b > 5))
        .group_by(_.c)
        .having(_.a.mean() > 0.0)
        .aggregate(a_mean=_.a.mean(), b_sum=_.b.sum())
        .order_by(_.a_mean)
        .mutate(
            arrays=ibis.array([1, 2, 3]),
            maps=ibis.map({"a": 1, "b": 2}),
            structs=ibis.struct({"a": [1, 2, 3], "b": {"c": 1, "d": 2}}),
        )
    )

    expr.visualize(
        verbose=args.verbose > 0,
        label_edges=args.label_edges,
        node_attr=args.node_attr,
        edge_attr=args.edge_attr,
    )



from __future__ import annotations

import math
from functools import partial, reduce
from typing import TYPE_CHECKING, Any

import sqlglot as sg
import sqlglot.expressions as sge
from sqlglot.dialects import DuckDB

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends.sql.compilers.base import (
    NULL,
    STAR,
    AggGen,
    SQLGlotCompiler,
)
from xorq.vendor.ibis.backends.sql.datatypes import DuckDBType
from xorq.vendor.ibis.util import gen_name


if TYPE_CHECKING:
    from collections.abc import Mapping

    import xorq.vendor.ibis.expr.types as ir


_INTERVAL_SUFFIXES = {
    "ms": "milliseconds",
    "us": "microseconds",
    "s": "seconds",
    "m": "minutes",
    "h": "hours",
    "D": "days",
    "M": "months",
    "Y": "years",
}


class DuckDBCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = DuckDB
    type_mapper = DuckDBType

    agg = AggGen(supports_filter=True, supports_order_by=True)

    supports_qualify = True

    LOWERED_OPS = {
        ops.Sample: None,
        ops.StringSlice: None,
    }

    SIMPLE_OPS = {
        ops.Arbitrary: "any_value",
        ops.ArrayMin: "list_min",
        ops.ArrayMax: "list_max",
        ops.ArrayAny: "list_bool_or",
        ops.ArrayAll: "list_bool_and",
        ops.ArraySum: "list_sum",
        ops.ArrayMean: "list_avg",
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.EndsWith: "suffix",
        ops.ExtractIsoYear: "isoyear",
        ops.IntegerRange: "range",
        ops.TimestampRange: "range",
        ops.MapLength: "cardinality",
        ops.Mode: "mode",
        ops.TimeFromHMS: "make_time",
        ops.GeoPoint: "st_point",
        ops.GeoAsText: "st_astext",
        ops.GeoArea: "st_area",
        ops.GeoBuffer: "st_buffer",
        ops.GeoCentroid: "st_centroid",
        ops.GeoContains: "st_contains",
        ops.GeoCovers: "st_covers",
        ops.GeoCoveredBy: "st_coveredby",
        ops.GeoCrosses: "st_crosses",
        ops.GeoDifference: "st_difference",
        ops.GeoDisjoint: "st_disjoint",
        ops.GeoDistance: "st_distance",
        ops.GeoDWithin: "st_dwithin",
        ops.GeoEndPoint: "st_endpoint",
        ops.GeoEnvelope: "st_envelope",
        ops.GeoEquals: "st_equals",
        ops.GeoFlipCoordinates: "st_flipcoordinates",
        ops.GeoGeometryType: "st_geometrytype",
        ops.GeoIntersection: "st_intersection",
        ops.GeoIntersects: "st_intersects",
        ops.GeoIsValid: "st_isvalid",
        ops.GeoLength: "st_length",
        ops.GeoNPoints: "st_npoints",
        ops.GeoOverlaps: "st_overlaps",
        ops.GeoStartPoint: "st_startpoint",
        ops.GeoTouches: "st_touches",
        ops.GeoUnion: "st_union",
        ops.GeoUnaryUnion: "st_union_agg",
        ops.GeoWithin: "st_within",
        ops.GeoX: "st_x",
        ops.GeoY: "st_y",
    }

    def to_sqlglot(
        self,
        expr: ir.Expr,
        *,
        limit: str | None = None,
        params: Mapping[ir.Expr, Any] | None = None,
    ):
        sql = super().to_sqlglot(expr, limit=limit, params=params)

        table_expr = expr.as_table()
        geocols = table_expr.schema().geospatial

        if not geocols:
            return sql

        quoted = self.quoted
        return sg.select(
            sge.Star(
                replace=[
                    self.f.st_aswkb(sg.column(col, quoted=quoted)).as_(
                        col, quoted=quoted
                    )
                    for col in geocols
                ]
            )
        ).from_(sql.subquery())

    def visit_StructColumn(self, op, *, names, values):
        return sge.Struct.from_arg_list(
            [
                sge.PropertyEQ(
                    this=sg.to_identifier(name, quoted=self.quoted), expression=value
                )
                for name, value in zip(names, values)
            ]
        )

    def visit_ArrayDistinct(self, op, *, arg):
        return self.if_(
            arg.is_(NULL),
            NULL,
            self.f.list_distinct(arg)
            + self.if_(
                self.f.list_count(arg) < self.f.len(arg),
                self.f.array(NULL),
                self.f.array(),
            ),
        )

    def visit_ArrayPosition(self, op, *, arg, other):
        return self.if_(
            arg.is_(NULL) | other.is_(NULL),
            NULL,
            self.f.coalesce(self.f.list_indexof(arg, other), 0),
        )

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.array_agg(arg, where=where, order_by=order_by)

    def visit_ArrayIndex(self, op, *, arg, index):
        return self.f.list_extract(arg, index + self.cast(index >= 0, op.index.dtype))

    def visit_ArrayRepeat(self, op, *, arg, times):
        func = sge.Lambda(this=arg, expressions=[sg.to_identifier("_")])
        return self.f.flatten(self.f.list_apply(self.f.range(times), func))

    # TODO(kszucs): this could be moved to the base SQLGlotCompiler
    def visit_Sample(
        self, op, *, parent, fraction: float, method: str, seed: int | None, **_
    ):
        sample = sge.TableSample(
            method="bernoulli" if method == "row" else "system",
            percent=sge.convert(fraction * 100.0),
            seed=None if seed is None else sge.convert(seed),
        )

        return self._make_sample_backwards_compatible(sample=sample, parent=parent)

    def visit_ArraySlice(self, op, *, arg, start, stop):
        arg_length = self.f.len(arg)

        if start is None:
            start = 0
        else:
            start = self.f.least(arg_length, self._neg_idx_to_pos(arg, start))

        if stop is None:
            stop = arg_length
        else:
            stop = self._neg_idx_to_pos(arg, stop)

        return self.f.list_slice(arg, start + 1, stop)

    def visit_ArrayMap(self, op, *, arg, body, param):
        lamduh = sge.Lambda(this=body, expressions=[sg.to_identifier(param)])
        return self.f.list_apply(arg, lamduh)

    def visit_ArrayFilter(self, op, *, arg, body, param):
        lamduh = sge.Lambda(this=body, expressions=[sg.to_identifier(param)])
        return self.f.list_filter(arg, lamduh)

    def visit_ArrayIntersect(self, op, *, left, right):
        param = sg.to_identifier("x")
        body = self.f.list_contains(right, param)
        lamduh = sge.Lambda(this=body, expressions=[param])
        return self.f.list_filter(left, lamduh)

    def visit_ArrayRemove(self, op, *, arg, other):
        x = sg.to_identifier(util.gen_name("x"))
        should_keep_null = sg.and_(x.is_(NULL), other.is_(sg.not_(NULL)))
        cond = sg.or_(x.neq(other), should_keep_null)
        return self.f.list_filter(arg, sge.Lambda(this=cond, expressions=[x]))

    def visit_ArrayUnion(self, op, *, left, right):
        arg = self.f.list_concat(left, right)
        return self.if_(
            arg.is_(NULL),
            NULL,
            self.f.list_distinct(arg)
            + self.if_(
                self.f.list_count(arg) < self.f.len(arg),
                self.f.array(NULL),
                self.f.array(),
            ),
        )

    def visit_ArrayZip(self, op, *, arg):
        i = sg.to_identifier("i")
        body = sge.Struct.from_arg_list(
            [
                sge.PropertyEQ(this=k, expression=v[i])
                for k, v in zip(map(sge.convert, op.dtype.value_type.names), arg)
            ]
        )
        func = sge.Lambda(this=body, expressions=[i])
        zipped_arrays = self.f.list_apply(
            self.f.range(
                1,
                # DuckDB Range excludes upper bound
                self.f.greatest(*map(self.f.len, arg)) + 1,
            ),
            func,
        )
        # if any of the input arrays in arg are NULL, the result is NULL
        any_arg_null = sg.or_(*(arr.is_(NULL) for arr in arg))
        return self.if_(any_arg_null, NULL, zipped_arrays)

    def visit_Array(self, op, *, exprs):
        return self.cast(self.f.array(*exprs), op.dtype)

    def visit_Map(self, op, *, keys, values):
        # workaround for https://github.com/ibis-project/ibis/issues/8632
        return self.if_(
            sg.or_(keys.is_(NULL), values.is_(NULL)),
            NULL,
            self.f.map(
                self.cast(keys, op.keys.dtype), self.cast(values, op.values.dtype)
            ),
        )

    def visit_MapGet(self, op, *, arg, key, default):
        return self.if_(
            arg.is_(NULL),
            NULL,
            self.f.ifnull(
                self.f.list_extract(
                    self.if_(key.is_(NULL), NULL, self.f.element_at(arg, key)), 1
                ),
                default,
            ),
        )

    def visit_MapContains(self, op, *, arg, key):
        return self.if_(
            arg.is_(NULL),
            NULL,
            self.f.len(self.if_(key.is_(NULL), NULL, self.f.element_at(arg, key))).neq(
                0
            ),
        )

    def visit_MapKeys(self, op, *, arg):
        return self.if_(arg.is_(NULL), NULL, self.f.map_keys(arg))

    def visit_MapValues(self, op, *, arg):
        return self.if_(arg.is_(NULL), NULL, self.f.map_values(arg))

    def visit_MapMerge(self, op, *, left, right):
        return self.if_(
            sg.or_(left.is_(NULL), right.is_(NULL)),
            NULL,
            self.f.map_concat(left, right),
        )

    def visit_ToJSONMap(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).eq(sge.convert("OBJECT")),
            self.cast(self.cast(arg, dt.json), op.dtype),
            NULL,
        )

    def visit_ToJSONArray(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).eq(sge.convert("ARRAY")),
            self.cast(self.cast(arg, dt.json), op.dtype),
            NULL,
        )

    def visit_UnwrapJSONString(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).eq(sge.convert("VARCHAR")),
            self.f.json_extract_string(arg, "$"),
            NULL,
        )

    def visit_UnwrapJSONInt64(self, op, *, arg):
        arg_type = self.f.json_type(arg)
        return self.if_(
            arg_type.isin(sge.convert("UBIGINT"), sge.convert("BIGINT")),
            self.cast(arg, op.dtype),
            NULL,
        )

    def visit_UnwrapJSONFloat64(self, op, *, arg):
        arg_type = self.f.json_type(arg)
        return self.if_(
            arg_type.isin(
                sge.convert("UBIGINT"), sge.convert("BIGINT"), sge.convert("DOUBLE")
            ),
            self.cast(arg, op.dtype),
            NULL,
        )

    def visit_UnwrapJSONBoolean(self, op, *, arg):
        return self.if_(
            self.f.json_type(arg).eq(sge.convert("BOOLEAN")),
            self.cast(arg, op.dtype),
            NULL,
        )

    def visit_ArrayConcat(self, op, *, arg):
        # TODO(cpcloud): map ArrayConcat to this in sqlglot instead of here
        return reduce(self.f.list_concat, arg)

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        if unit.short == "ns":
            raise com.UnsupportedOperationError(
                f"{self.dialect} doesn't support nanosecond interval resolutions"
            )

        if unit.singular == "week":
            return self.f.to_days(arg * 7)
        return self.f[f"to_{unit.plural}"](arg)

    def visit_FindInSet(self, op, *, needle, values):
        return self.if_(
            needle.is_(NULL),
            NULL,
            self.f.coalesce(self.f.list_indexof(self.f.array(*values), needle), 0),
        )

    def visit_CountDistinctStar(self, op, *, where, arg):
        # use a tuple because duckdb doesn't accept COUNT(DISTINCT a, b, c, ...)
        #
        # this turns the expression into COUNT(DISTINCT (a, b, c, ...))
        row = sge.Tuple(
            expressions=list(
                map(partial(sg.column, quoted=self.quoted), op.arg.schema.keys())
            )
        )
        return self.agg.count(sge.Distinct(expressions=[row]), where=where)

    def visit_ExtractMillisecond(self, op, *, arg):
        return self.f.mod(self.f.extract("ms", arg), 1_000)

    # DuckDB extracts subminute microseconds and milliseconds
    # so we have to finesse it a little bit
    def visit_ExtractMicrosecond(self, op, *, arg):
        return self.f.mod(self.f.extract("us", arg), 1_000_000)

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        unit = unit.short
        if unit == "ms":
            return self.f.epoch_ms(arg)
        elif unit == "s":
            return sge.UnixToTime(this=arg)
        else:
            raise com.UnsupportedOperationError(f"{unit!r} unit is not supported!")

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds, **_
    ):
        args = [year, month, day, hours, minutes, seconds]

        func = "make_timestamp"
        if (timezone := op.dtype.timezone) is not None:
            func += "tz"
            args.append(timezone)

        return self.f[func](*args)

    def visit_Cast(self, op, *, arg, to):
        if to.is_interval():
            func = self.f[f"to_{_INTERVAL_SUFFIXES[to.unit.short]}"]
            return func(sg.cast(arg, to=self.type_mapper.from_ibis(dt.int32)))
        elif to.is_timestamp() and op.arg.dtype.is_integer():
            return self.f.to_timestamp(arg)
        elif to.is_geospatial() and op.arg.dtype.is_binary():
            return self.f.st_geomfromwkb(arg)

        return self.cast(arg, to)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_interval():
            if dtype.unit.short == "ns":
                raise com.UnsupportedOperationError(
                    f"{self.dialect} doesn't support nanosecond interval resolutions"
                )

            return sge.Interval(
                this=sge.convert(str(value)), unit=dtype.resolution.upper()
            )
        elif dtype.is_uuid():
            return self.cast(str(value), dtype)
        elif dtype.is_binary():
            return self.cast("".join(map("\\x{:02x}".format, value)), dtype)
        elif dtype.is_numeric():
            # cast non finite values to float because that's the behavior of
            # duckdb when a mixed decimal/float operation is performed
            #
            # float will be upcast to double if necessary by duckdb
            if not math.isfinite(value):
                return self.cast(
                    str(value), to=dt.float32 if dtype.is_decimal() else dtype
                )
            if dtype.is_floating() or dtype.is_integer():
                return sge.convert(value)
            return self.cast(value, dtype)
        elif dtype.is_time():
            return self.f.make_time(
                value.hour, value.minute, value.second + value.microsecond / 1e6
            )
        elif dtype.is_timestamp():
            args = [
                value.year,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second + value.microsecond / 1e6,
            ]

            funcname = "make_timestamp"

            if (tz := dtype.timezone) is not None:
                funcname += "tz"
                args.append(tz)

            return self.f[funcname](*args)
        elif dtype.is_struct():
            return self.cast(
                sge.Struct.from_arg_list(
                    [
                        self.visit_Literal(
                            ops.Literal(v, field_dtype), value=v, dtype=field_dtype
                        )
                        for field_dtype, v in zip(dtype.types, value.values())
                    ]
                ),
                op.dtype,
            )
        else:
            return None

    def _neg_idx_to_pos(self, array, idx):
        arg_length = self.f.array_size(array)
        return self.if_(
            idx >= 0,
            idx,
            # Need to have the greatest here to handle the case where
            # abs(neg_index) > arg_length
            # e.g. where the magnitude of the negative index is greater than the
            # length of the array
            # You cannot index a[:-3] if a = [1, 2]
            arg_length + self.f.greatest(idx, -arg_length),
        )

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "sample":
            raise com.UnsupportedOperationError(
                f"{self.dialect} only implements `pop` correlation coefficient"
            )

        # TODO: rewrite rule?
        if (left_type := op.left.dtype).is_boolean():
            left = self.cast(left, dt.Int32(nullable=left_type.nullable))

        if (right_type := op.right.dtype).is_boolean():
            right = self.cast(right, dt.Int32(nullable=right_type.nullable))

        return self.agg.corr(left, right, where=where)

    def visit_GeoConvert(self, op, *, arg, source, target):
        # 4th argument is to specify that the result is always_xy so that it
        # matches the behavior of the equivalent geopandas functionality
        return self.f.st_transform(arg, source, target, True)

    def visit_TimestampNow(self, op):
        """DuckDB current timestamp defaults to timestamp + tz."""
        return self.cast(super().visit_TimestampNow(op), dt.timestamp)

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        return self.f.regexp_extract(arg, pattern, index, dialect=self.dialect)

    def visit_RegexReplace(self, op, *, arg, pattern, replacement):
        return self.f.regexp_replace(
            arg, pattern, replacement, "g", dialect=self.dialect
        )

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.first(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if not include_null:
            cond = arg.is_(sg.not_(NULL, copy=False))
            where = cond if where is None else sge.And(this=cond, expression=where)
        return self.agg.last(arg, where=where, order_by=order_by)

    def visit_Quantile(self, op, *, arg, quantile, where):
        suffix = "cont" if op.arg.dtype.is_numeric() else "disc"
        funcname = f"percentile_{suffix}"
        return self.agg[funcname](arg, quantile, where=where)

    def visit_MultiQuantile(self, op, *, arg, quantile, where):
        return self.visit_Quantile(op, arg=arg, quantile=quantile, where=where)

    def visit_ApproxQuantile(self, op, *, arg, quantile, where):
        if not op.arg.dtype.is_floating():
            arg = self.cast(arg, dt.float64)
        return self.agg.approx_quantile(arg, quantile, where=where)

    visit_ApproxMultiQuantile = visit_ApproxQuantile

    def visit_HexDigest(self, op, *, arg, how):
        if how in ("md5", "sha256"):
            return getattr(self.f, how)(arg)
        else:
            raise NotImplementedError(f"No available hashing function for {how}")

    def visit_Hash(self, op, *, arg):
        # duckdb's hash() returns a uint64, but ops.Hash is supposed to be int64
        # So do HASH(x)::BITSTRING::BIGINT
        raw = self.f.hash(arg)
        bitstring = sg.cast(sge.convert(raw), to=sge.DataType.Type.BIT, copy=False)
        int64 = sg.cast(bitstring, to=sge.DataType.Type.BIGINT, copy=False)
        return int64

    def visit_StringConcat(self, op, *, arg):
        return reduce(lambda x, y: sge.DPipe(this=x, expression=y), arg)

    def visit_StringSlice(self, op, *, arg, start, end):
        if start is not None:
            start += 1
        # workaround for https://github.com/duckdb/duckdb/issues/11431
        start = self.f.ifnull(start, 1)
        end = self.f.ifnull(end, -1)
        return self.f.array_slice(arg, start, end)

    def visit_StructField(self, op, *, arg, field):
        if not isinstance(op.arg, (ops.Field, sge.Struct)):
            # parenthesize anything that isn't a simple field access
            return sge.Dot(
                this=sge.paren(arg),
                expression=sg.to_identifier(field, quoted=self.quoted),
            )
        return super().visit_StructField(op, arg=arg, field=field)

    def visit_RandomScalar(self, op, **kwargs):
        return self.f.random()

    def visit_RandomUUID(self, op, **kwargs):
        return self.f.uuid()

    def visit_TypeOf(self, op, *, arg):
        return self.f.coalesce(self.f.nullif(self.f.typeof(arg), '"NULL"'), "NULL")

    def visit_DropColumns(self, op, *, parent, columns_to_drop):
        quoted = self.quoted
        # duckdb doesn't support specifying the table name of the column name
        # to drop, e.g., in SELECT t.* EXCLUDE (t.a) FROM t, the t.a bit
        #
        # technically it's not necessary, here's why
        #
        # if the table is specified then it's unambiguous when there are overlapping
        # column names, say, from a join, for example
        # (assuming t and s both have a column named `a`)
        #
        # SELECT t.* EXCLUDE (a), s.* FROM t JOIN s ON id
        #
        # This would exclude t.a and include s.a
        #
        # if it's a naked star projection from a join, like
        #
        # SELECT * EXCLUDE (a) FROM t JOIN s ON id
        #
        # then that means "exclude all columns named `a`"
        excludes = [sg.column(column, quoted=quoted) for column in columns_to_drop]
        star = sge.Star(**{"except": excludes})
        table = sg.to_identifier(parent.alias_or_name, quoted=quoted)
        column = sge.Column(this=star, table=table)
        return sg.select(column).from_(parent)

    def visit_TableUnnest(
        self,
        op,
        *,
        parent,
        column,
        column_name: str,
        offset: str | None,
        keep_empty: bool,
    ):
        quoted = self.quoted

        column_alias = sg.to_identifier(gen_name("table_unnest_column"), quoted=quoted)

        overlaps_with_parent = column_name in op.parent.schema
        computed_column = column_alias.as_(column_name, quoted=quoted)

        selcols = []

        table = sg.to_identifier(parent.alias_or_name, quoted=quoted)

        if offset is not None:
            # TODO: clean this up once WITH ORDINALITY is supported in DuckDB
            # no need for struct_extract once that's upstream
            column = self.f.list_zip(column, self.f.range(self.f.len(column)))
            extract = self.f.struct_extract(column_alias, 1).as_(
                column_name, quoted=quoted
            )

            if overlaps_with_parent:
                replace = sge.Column(this=sge.Star(replace=[extract]), table=table)
                selcols.append(replace)
            else:
                selcols.append(sge.Column(this=STAR, table=table))
                selcols.append(extract)

            selcols.append(
                self.f.struct_extract(column_alias, 2).as_(offset, quoted=quoted)
            )
        elif overlaps_with_parent:
            selcols.append(
                sge.Column(this=sge.Star(replace=[computed_column]), table=table)
            )
        else:
            selcols.append(sge.Column(this=STAR, table=table))
            selcols.append(computed_column)

        unnest = sge.Unnest(
            expressions=[column],
            alias=sge.TableAlias(
                this=sg.to_identifier(gen_name("table_unnest"), quoted=quoted),
                columns=[column_alias],
            ),
        )
        return (
            sg.select(*selcols)
            .from_(parent)
            .join(unnest, join_type="CROSS" if not keep_empty else "LEFT")
        )


compiler = DuckDBCompiler()



import numpy as np
import pandas as pd
import pytest

import xorq.api as xo


QUOTES_TABLE_NAME = "quotes"


@pytest.fixture(scope="session")
def quotes_df():
    # Quote data (more frequent updates)
    quote_dates = pd.date_range(
        start="2024-01-01", end="2024-01-31", freq="30min"
    ).astype("datetime64[us]")

    quotes = pd.DataFrame(
        {
            "timestamp": quote_dates,
            "symbol": np.random.choice(["AAPL", "GOOGL", "MSFT"], len(quote_dates)),
            "bid": np.random.uniform(99, 198, len(quote_dates)),
            "ask": np.random.uniform(101, 202, len(quote_dates)),
        }
    )

    return quotes


@pytest.fixture
def trades_df():
    # Create sample trading data
    np.random.seed(42)
    dates = pd.date_range(start="2024-01-01", end="2024-01-31", freq="h").astype(
        "datetime64[us]"
    )

    # Trades data
    trades = pd.DataFrame(
        {
            "timestamp": dates,
            "symbol": np.random.choice(["AAPL", "GOOGL", "MSFT"], len(dates)),
            "price": np.random.uniform(100, 200, len(dates)),
            "volume": np.random.randint(100, 1000, len(dates)),
        }
    )

    return trades


@pytest.fixture(scope="session")
def iceberg_con(tmp_path_factory, quotes_df):
    warehouse_path = str(tmp_path_factory.mktemp("warehouse"))
    con = xo.pyiceberg.connect(warehouse_path=warehouse_path)
    con.create_table(QUOTES_TABLE_NAME, quotes_df, overwrite=True)
    return con


@pytest.fixture(scope="session")
def quotes_table(iceberg_con):
    return iceberg_con.table(QUOTES_TABLE_NAME)



from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.rules as rlz
import xorq.vendor.ibis.expr.types as ir
from xorq.common.exceptions import XorqInputError
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.common.annotations import annotated, attribute
from xorq.vendor.ibis.common.deferred import Deferred, Resolver, deferrable
from xorq.vendor.ibis.common.grounds import Concrete
from xorq.vendor.ibis.common.selectors import Selector  # noqa: TCH001
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001


if TYPE_CHECKING:
    from typing_extensions import Self


class Builder(Concrete):
    pass


@deferrable(repr="<case>")
def _finish_searched_case(cases, results, default) -> ir.Value:
    """Finish constructing a SearchedCase expression.

    This is split out into a separate function to allow for deferred arguments
    to resolve.
    """
    return ops.SearchedCase(cases=cases, results=results, default=default).to_expr()


class SearchedCaseBuilder(Builder):
    """A case builder, used for constructing `ibis.case()` expressions."""

    cases: VarTuple[Union[Resolver, ops.Value[dt.Boolean]]] = ()
    results: VarTuple[Union[Resolver, ops.Value]] = ()
    default: Optional[Union[Resolver, ops.Value]] = None

    def when(self, case_expr: Any, result_expr: Any) -> Self:
        """Add a new condition and result to the `CASE` expression.

        Parameters
        ----------
        case_expr
            Predicate expression to use for this case.
        result_expr
            Value when the case predicate evaluates to true.

        """
        return self.copy(
            cases=self.cases + (case_expr,), results=self.results + (result_expr,)
        )

    def else_(self, result_expr: Any) -> Self:
        """Add a default value for the `CASE` expression.

        Parameters
        ----------
        result_expr
            Value to use when all case predicates evaluate to false.

        """
        return self.copy(default=result_expr)

    def end(self) -> ir.Value | Deferred:
        """Finish the `CASE` expression."""
        return _finish_searched_case(self.cases, self.results, self.default)


class SimpleCaseBuilder(Builder):
    """A case builder, used for constructing `Column.case()` expressions."""

    base: ops.Value
    cases: VarTuple[ops.Value] = ()
    results: VarTuple[ops.Value] = ()
    default: Optional[ops.Value] = None

    def when(self, case_expr: Any, result_expr: Any) -> Self:
        """Add a new condition and result to the `CASE` expression.

        Parameters
        ----------
        case_expr
            Expression to equality-compare with base expression. Must be
            comparable with the base.
        result_expr
            Value when the case predicate evaluates to true.

        """
        if not isinstance(case_expr, ir.Value):
            case_expr = ibis.literal(case_expr)
        if not isinstance(result_expr, ir.Value):
            result_expr = ibis.literal(result_expr)
        return self.copy(
            cases=self.cases + (case_expr,), results=self.results + (result_expr,)
        )

    def else_(self, result_expr: Any) -> Self:
        """Add a default value for the `CASE` expression.

        Parameters
        ----------
        result_expr
            Value to use when all case predicates evaluate to false.

        """
        return self.copy(default=result_expr)

    def end(self) -> ir.Value:
        """Finish the `CASE` expression."""
        if (default := self.default) is None:
            default = ibis.null().cast(rlz.highest_precedence_dtype(self.results))
        return ops.SimpleCase(
            cases=self.cases, results=self.results, default=default, base=self.base
        ).to_expr()


RowsWindowBoundary = ops.WindowBoundary[dt.Integer]
RangeWindowBoundary = ops.WindowBoundary[dt.Numeric | dt.Interval]


class WindowBuilder(Builder):
    """An unbound window frame specification.

    Notes
    -----
    This class is patterned after SQL window frame clauses.

    Using `None` for `preceding` or `following` indicates an unbounded frame.

    Use 0 for `CURRENT ROW`.

    """

    how: Literal["rows", "range"] = "rows"
    start: Optional[RangeWindowBoundary] = None
    end: Optional[RangeWindowBoundary] = None
    groupings: VarTuple[Union[str, Resolver, Selector, ops.Value]] = ()
    orderings: VarTuple[Union[str, Resolver, Selector, ops.SortKey]] = ()

    @attribute
    def _table(self):
        inputs = (
            self.start,
            self.end,
            *self.groupings,
            *self.orderings,
        )
        valuerels = (v.relations for v in inputs if isinstance(v, ops.Value))
        relations = frozenset().union(*valuerels)
        if len(relations) == 0:
            return None
        elif len(relations) == 1:
            (table,) = relations
            return table
        else:
            raise XorqInputError("Window frame can only depend on a single relation")

    def _maybe_cast_boundaries(self, start, end):
        if start and end:
            if start.dtype.is_interval() and end.dtype.is_numeric():
                return start, ops.Cast(end.value, start.dtype)
            elif start.dtype.is_numeric() and end.dtype.is_interval():
                return ops.Cast(start.value, end.dtype), end
        return start, end

    def _determine_how(self, start, end):
        if start and not start.dtype.is_integer():
            return self.range
        elif end and not end.dtype.is_integer():
            return self.range
        else:
            return self.rows

    def _validate_boundaries(self, start, end):
        start_, end_ = -math.inf, math.inf
        if start and isinstance(lit := start.value, ops.Literal):
            start_ = -lit.value if start.preceding else lit.value
        if end and isinstance(lit := end.value, ops.Literal):
            end_ = -lit.value if end.preceding else lit.value

        if start_ > end_:
            raise XorqInputError(
                "Window frame's start point must be greater than its end point"
            )

    @annotated
    def rows(
        self, start: Optional[RowsWindowBoundary], end: Optional[RowsWindowBoundary]
    ):
        self._validate_boundaries(start, end)
        start, end = self._maybe_cast_boundaries(start, end)
        return self.copy(how="rows", start=start, end=end)

    @annotated
    def range(
        self, start: Optional[RangeWindowBoundary], end: Optional[RangeWindowBoundary]
    ):
        self._validate_boundaries(start, end)
        start, end = self._maybe_cast_boundaries(start, end)
        return self.copy(how="range", start=start, end=end)

    @annotated
    def between(
        self, start: Optional[RangeWindowBoundary], end: Optional[RangeWindowBoundary]
    ):
        self._validate_boundaries(start, end)
        start, end = self._maybe_cast_boundaries(start, end)
        method = self._determine_how(start, end)
        return method(start, end)

    def group_by(self, expr) -> Self:
        return self.copy(groupings=self.groupings + util.promote_tuple(expr))

    def order_by(self, expr) -> Self:
        return self.copy(orderings=self.orderings + util.promote_tuple(expr))

    def bind(self, table):
        if table is None:
            if self._table is None:
                raise XorqInputError("Cannot bind window frame without a table")
            else:
                table = self._table.to_expr()

        return self.copy(
            groupings=table.bind(self.groupings), orderings=table.bind(self.orderings)
        )


class LegacyWindowBuilder(WindowBuilder):
    def _is_negative(self, value):
        if value is None:
            return False
        if isinstance(value, ir.Scalar):
            value = value.op().value
        return value < 0

    def preceding_following(self, preceding, following, how=None) -> Self:
        preceding_tuple = has_preceding = False
        following_tuple = has_following = False
        if preceding is not None:
            preceding_tuple = isinstance(preceding, tuple)
            has_preceding = True
        if following is not None:
            following_tuple = isinstance(following, tuple)
            has_following = True

        if (preceding_tuple and has_following) or (following_tuple and has_preceding):
            raise XorqInputError(
                "Can only specify one window side when you want an off-center window"
            )
        elif preceding_tuple:
            start, end = preceding
            if end is None:
                raise XorqInputError("preceding end point cannot be None")
            elif self._is_negative(end):
                raise XorqInputError("preceding end point must be non-negative")
            elif self._is_negative(start):
                raise XorqInputError("preceding start point must be non-negative")
            between = (
                None if start is None else ops.WindowBoundary(start, preceding=True),
                ops.WindowBoundary(end, preceding=True),
            )
        elif following_tuple:
            start, end = following
            if start is None:
                raise XorqInputError("following start point cannot be None")
            elif self._is_negative(start):
                raise XorqInputError("following start point must be non-negative")
            elif self._is_negative(end):
                raise XorqInputError("following end point must be non-negative")
            between = (
                ops.WindowBoundary(start, preceding=False),
                None if end is None else ops.WindowBoundary(end, preceding=False),
            )
        elif has_preceding and has_following:
            between = (
                ops.WindowBoundary(preceding, preceding=True),
                ops.WindowBoundary(following, preceding=False),
            )
        elif has_preceding:
            if self._is_negative(preceding):
                raise XorqInputError("preceding end point must be non-negative")
            between = (ops.WindowBoundary(preceding, preceding=True), None)
        elif has_following:
            if self._is_negative(following):
                raise XorqInputError("following end point must be non-negative")
            between = (None, ops.WindowBoundary(following, preceding=False))

        if how is None:
            return self.between(*between)
        elif how == "rows":
            return self.rows(*between)
        elif how == "range":
            return self.range(*between)
        else:
            raise ValueError(f"Invalid window frame type: {how}")



import importlib

import pandas as pd
import pytest
import toolz

from xorq.backends.datafusion import Backend as DatafusionBackend
from xorq.backends.postgres import Backend as PostgresBackend
from xorq.backends.snowflake import Backend as SnowflakeBackend
from xorq.backends.snowflake import SnowflakeAuthenticator
from xorq.backends.sqlite import Backend as SqliteBackend
from xorq.common.utils.env_utils import maybe_substitute_env_var
from xorq.common.utils.ibis_utils import (
    from_ibis,
    ibis_utils_backend_registry_context,
    map_ibis,
)


ibis = pytest.importorskip("ibis")


def connect_postgres():
    from xorq.common.utils.postgres_utils import (
        make_connection_defaults,
        make_credential_defaults,
    )

    return ibis.postgres.connect(
        **toolz.valmap(
            maybe_substitute_env_var,
            (make_credential_defaults() | make_connection_defaults()),
        )
    )


def connect_snowflake():
    from xorq.common.utils.snowflake_keypair_utils import maybe_decrypt_private_key
    from xorq.common.utils.snowflake_utils import (
        make_connection_defaults,
        make_credential_defaults,
    )

    database = "SNOWFLAKE_SAMPLE_DATA"
    schema = "TPCH_SF1"

    connection_defaults = toolz.valmap(
        maybe_substitute_env_var,
        (
            make_credential_defaults(authenticator=SnowflakeAuthenticator.keypair)
            | make_connection_defaults()
        ),
    )

    if "private_key" in connection_defaults:
        connection_defaults = maybe_decrypt_private_key(connection_defaults)

    return ibis.snowflake.connect(
        database=f"{database}/{schema}",
        **connection_defaults,
    )


@pytest.mark.parametrize(
    "get_con,backend_type",
    (
        pytest.param(
            connect_postgres,
            PostgresBackend,
            marks=[
                pytest.mark.skipif(
                    importlib.util.find_spec("psycopg2") is None,
                    reason="requires backwards postgres compatibility",
                )
            ],
            id="postgres",
        ),
        pytest.param(
            lambda: ibis.datafusion.connect(), DatafusionBackend, id="datafusion"
        ),
        pytest.param(lambda: ibis.sqlite.connect(), SqliteBackend, id="sqlite"),
    ),
)
def test_backends(get_con, backend_type):
    conn = get_con()
    ta = conn.create_table("foo", pd.DataFrame({"id": [1, 2, 3]}))
    expr = ta.select(ibis._.id).filter(ibis._.id > 2)

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None
    assert not xorq_expr.execute().empty
    assert isinstance(xorq_expr._find_backend(), backend_type)

    try:
        conn.drop_table("foo")
    except Exception as e:
        print(e)


@pytest.mark.skipif(
    importlib.util.find_spec("snowflake") is None, reason="requires snowflake"
)
@pytest.mark.snowflake
def test_snowflake_backend():
    snow_conn = connect_snowflake()
    assert snow_conn.list_tables() is not None
    with ibis_utils_backend_registry_context():
        xorq_con = map_ibis(snow_conn, None)
    assert isinstance(xorq_con, SnowflakeBackend)
    assert xorq_con.list_tables() is not None



import pytest

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
    SourceCache,
)
from xorq.expr.ml.pipeline_lib import (
    Pipeline,
)


sklearn = pytest.importorskip("sklearn")
load_iris = sklearn.datasets.load_iris
train_test_split = sklearn.model_selection.train_test_split
KNeighborsClassifier = sklearn.neighbors.KNeighborsClassifier
StandardScaler = sklearn.preprocessing.StandardScaler


def make_pipeline():
    clf = sklearn.pipeline.Pipeline(
        steps=[
            ("scaler", StandardScaler()),
            ("knn", KNeighborsClassifier(n_neighbors=11)),
        ]
    )
    return clf


def make_data():
    iris = load_iris(as_frame=True)
    X = iris.data[["sepal length (cm)", "sepal width (cm)"]].rename(
        columns={
            "sepal length (cm)": "sepal_length",
            "sepal width (cm)": "sepal_width",
        }
    )
    y = iris.target
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, stratify=y, random_state=0
    )
    return X_train, X_test, y_train, y_test


def make_exprs(X_train, X_test, y_train, y_test):
    con = xo.connect()
    features = tuple(X_train.columns)
    target = y_train.name
    train = con.register(X_train.assign(**{target: y_train}), "train")
    test = con.register(X_test.assign(**{target: y_test}), "test")
    return train, test, features, target


@pytest.mark.parametrize("cache_cls", (None, ParquetCache, SourceCache))
def test_fittedstep_model(cache_cls):
    cache = cache_cls.from_kwargs() if cache_cls else cache_cls
    X_train, X_test, y_train, y_test = make_data()
    train, test, features, target = make_exprs(X_train, X_test, y_train, y_test)
    xorq_pipeline = Pipeline.from_instance(make_pipeline())
    fitted_pipeline = xorq_pipeline.fit(
        train, features=features, target=target, cache=cache
    )
    for fitted_step in (*fitted_pipeline.transform_steps, fitted_pipeline.predict_step):
        fitted_step.model



import operator
import pathlib

import pytest
import toolz

import xorq.api as xo
from xorq.expr.udf import make_pandas_udf


def pytest_collection_modifyitems(items):
    flight_tests_dir = pathlib.Path(__file__).parent
    for item in items:
        if pathlib.Path(item.fspath).is_relative_to(flight_tests_dir):
            item.add_marker(pytest.mark.xdist_group(name="flight"))


field_name = "price_per_carat"
schema = xo.schema({"sum_price": "float64", "sum_carat": "float64"})
return_type = xo.dtype("float64")


@pytest.fixture(scope="session")
def tls_kwargs():
    return {
        "ca_kwargs": {
            "common_name": "root_cert",
        },
        "server_kwargs": {
            "common_name": "server",
            "sans": ("localhost",),
        },
    }


@pytest.fixture(scope="session")
def mtls_kwargs(tls_kwargs):
    return tls_kwargs | {
        "client_kwargs": {
            "common_name": "client",
        },
    }


@make_pandas_udf(
    schema=schema,
    return_type=return_type,
    name="my_udf",
)
def my_udf(df):
    return df["sum_carat"].div(df["sum_price"])


def do_agg(expr):
    return (
        expr.group_by("cut")
        .agg(
            xo._.price.sum().name("sum_price"),
            xo._.carat.sum().name("sum_carat"),
            xo._.color.nunique().name("nunique_color"),
        )
        .cast({"sum_price": "float64"})
    )


my_udf_on_expr = toolz.compose(
    operator.methodcaller("name", field_name), my_udf.on_expr
)


@pytest.fixture(scope="function")
def con():
    return xo.connect()


@pytest.fixture(scope="function")
def diamonds(con):
    return xo.examples.diamonds.fetch(backend=con)


@pytest.fixture(scope="function")
def baseline(diamonds):
    expr = diamonds.pipe(do_agg).mutate(my_udf_on_expr).order_by("cut")
    return expr.execute()



import xorq.api as xo
from xorq.caching import SourceCache
from xorq.tests.util import assert_frame_equal


def test_read_parquet(persistent_sqlite_con, astronauts_parquet_path):
    t = persistent_sqlite_con.read_parquet(astronauts_parquet_path)
    assert not t.execute().empty


def test_can_be_cached(persistent_sqlite_con, astronauts_parquet_path):
    astronauts = persistent_sqlite_con.read_parquet(astronauts_parquet_path)
    expr = (
        astronauts.cache(SourceCache.from_kwargs(source=xo.duckdb.connect()))
        .filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
    )

    actual = expr.execute()
    expected = (
        astronauts.filter(xo._.number == 104)
        .select(xo._.id, xo._.number, xo._.nationwide_number, xo._.name)
        .execute()
    )

    assert_frame_equal(expected, actual)



import functools
import random
import socket
import time
from typing import Optional
from urllib.parse import urlunparse

from attrs import (
    define,
    field,
)
from attrs.validators import (
    in_,
    instance_of,
    optional,
)

from xorq.backends.xorq import connect as xo_connect
from xorq.common.utils.logging_utils import get_print_logger
from xorq.flight.action import AddExchangeAction
from xorq.flight.backend import Backend
from xorq.flight.server import (
    BasicAuthServerMiddlewareFactory,
    FlightServerDelegate,
    NoOpAuthHandler,
)


DEFAULT_AUTH_MIDDLEWARE = {
    "basic": BasicAuthServerMiddlewareFactory(
        {
            "test": "password",
        }
    )
}

allowed_schemes = ("grpc", "grpc+tls")
default_host = "localhost"


@define
class FlightUrl:
    scheme: str = field(default="grpc", validator=in_(allowed_schemes))
    host: str = field(default=default_host, validator=instance_of(str))
    username: Optional[str] = field(default=None, validator=optional(instance_of(str)))
    password: Optional[str] = field(default=None, validator=optional(instance_of(str)))
    port: Optional[int] = field(default=None, validator=optional(instance_of(int)))
    path: Optional[str] = field(default="", validator=instance_of(str))
    query: Optional[str] = field(default="", validator=instance_of(str))
    fragment: Optional[str] = field(default="", validator=instance_of(str))
    _socket = field(default=None, init=False)

    def __attrs_post_init__(self):
        if self.port is None:
            self.find_and_bind_socket()
        else:
            self.bind_socket()

    @property
    def client_kwargs(self):
        return {
            k: getattr(self, k)
            for k in (
                "host",
                "port",
                "username",
                "password",
            )
        }

    def find_and_bind_socket(self):
        while True:
            port = random.randint(5000, 8000)
            try:
                self._socket = self._bind_socket(self.host, port)
                self.port = port
                break
            except socket.error:
                pass

    def bind_socket(self):
        self._socket = self._bind_socket(self.host, self.port)

    def unbind_socket(self):
        self._socket = None

    def to_location(self):
        components = (
            self.scheme,
            f"{self.host}:{self.port}",
            self.path,
            "",
            self.query,
            self.fragment,
        )
        return str(urlunparse(components))

    @staticmethod
    def _bind_socket(host, port):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((host, port))
        return s

    @staticmethod
    def port_in_use(port, host="localhost"):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind((host, port))
                return False
            except socket.error:
                return True


class BasicAuth:
    def __init__(self, username, password):
        self.username = username
        self.password = password


def to_basic_auth_middleware(basic_auth: BasicAuth) -> dict:
    assert basic_auth is not None

    return {
        "basic": BasicAuthServerMiddlewareFactory(
            {
                basic_auth.username: basic_auth.password,
            }
        )
    }


class FlightServer:
    def __init__(
        self,
        flight_url=None,
        tls_certificates=(),  # list of (key, value) pairs
        verify_client=False,
        root_certificates=None,
        auth: BasicAuth = None,
        make_connection=xo_connect,
        exchangers=(),
    ):
        required_scheme = (
            "grpc+tls"
            if tls_certificates or verify_client or root_certificates
            else "grpc"
        )
        self.flight_url = flight_url or FlightUrl(scheme=required_scheme)
        assert self.flight_url.scheme == required_scheme
        self.tls_certificates = tls_certificates
        self.root_certificates = root_certificates
        self.auth = auth
        self.make_connection = make_connection
        self.verify_client = verify_client
        self.server = None
        self.exchangers = exchangers

    @classmethod
    def from_udxf(cls, expr, host=None, port=None, make_connection=None, **kwargs):
        from xorq.common.utils.graph_utils import walk_nodes
        from xorq.expr.relations import FlightUDXF

        # Find all FlightUDXF nodes in the expression
        exchangers = tuple({node.udxf for node in walk_nodes((FlightUDXF,), expr)})

        flight_url_kwargs = {
            key: value for key, value in (("host", host), ("port", port)) if value
        }
        flight_url = FlightUrl(**flight_url_kwargs)

        server_kwargs = {
            key: value
            for key, value in (
                ("exchangers", exchangers),
                ("make_connection", make_connection),
            )
            if value
        }

        return cls(
            flight_url,
            **kwargs | server_kwargs,
        )

    @property
    def auth_kwargs(self):
        kwargs = {
            "root_certificates": self.root_certificates,
            "tls_certificates": list(self.tls_certificates),
        }
        if self.auth:
            kwargs |= {
                "auth_handler": NoOpAuthHandler(),
                "middleware": to_basic_auth_middleware(self.auth),
            }
        return kwargs

    @property
    @functools.cache
    def con(self) -> Backend:
        kwargs = {
            "host": self.flight_url.host,
            "port": self.flight_url.port,
        }

        if self.auth is not None:
            kwargs["username"] = self.auth.username
            kwargs["password"] = self.auth.password

        if self.root_certificates:
            # this mapping is for Backend's FlightClient
            kwargs["tls_root_certs"] = self.root_certificates

        if self.tls_certificates:
            ((cert_chain, private_key), *rest) = self.tls_certificates
            if rest:
                # FIXME: what to do with multiple certs?
                raise ValueError(
                    "multiple TLS certificates provided; only one is supported"
                )
            kwargs["cert_chain"] = cert_chain
            kwargs["private_key"] = private_key

        instance = Backend()
        instance.do_connect(**kwargs)
        return instance

    @property
    def client(self):
        return self.con.con

    def __enter__(self):
        self.serve()
        return self

    def serve(self, block=False):
        self.flight_url.unbind_socket()
        self.server = FlightServerDelegate(
            self.make_connection,
            self.flight_url.to_location(),
            verify_client=self.verify_client,
            **self.auth_kwargs,
        )
        for udxf in self.exchangers:
            self.client.do_action(
                AddExchangeAction.name, udxf, options=self.client._options
            )
        if block:
            # https://arrow.apache.org/docs/python/generated/pyarrow.flight.FlightServerBase.html#pyarrow.flight.FlightServerBase.serve
            self.server.serve()

    def close(self, *args):
        args = args or (None, None, None)
        self.server.__exit__(*args)
        self.server = None
        self.flight_url.bind_socket()

    def wait(self):
        assert self.server
        while self.server:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                logger = get_print_logger()
                logger.info("Keyboard Interrupt: server shutting down")
                break

    def __exit__(self, *args):
        self.close(*args)


def connect(host=None, port=None, username=None, password=None, tls_kwargs=None):
    new_backend = Backend()
    new_backend.do_connect(
        **{
            k: v
            for k, v in (
                ("host", host),
                ("port", port),
                ("username", username),
                ("password", password),
            )
            if v
        },
        **(tls_kwargs.client_kwargs if tls_kwargs else {}),
    )

    return new_backend


__all__ = ["FlightServer", "FlightUrl", "BasicAuth", "connect"]



import pytest
import toolz

from xorq.common.utils.process_utils import Popened


make_bytes_contains = toolz.flip(bytes.__contains__)


def test_peek_line_until():
    popened = Popened(
        "sleep 1 && echo marker-text0 && sleep 1 && echo marker-text1",
        kwargs_tuple=(("shell", True),),
    )

    condition = make_bytes_contains(b"marker-text1")
    popened.stdout_peeker.peek_line_until(condition)


def test_peek_line_until_with_timeout():
    popened = Popened(
        "sleep 1 && echo marker-text0 && sleep 1 && echo marker-text1",
        kwargs_tuple=(("shell", True),),
    )

    marker_text = b"marker-text1"
    condition = toolz.flip(bytes.__contains__)(marker_text)
    popened.stdout_peeker.peek_line_until(condition, timeout=3)


def test_peek_line_until_with_timeout_raises():
    popened = Popened(
        "sleep 1 && echo marker-text0 && sleep 1 && echo marker-text1",
        kwargs_tuple=(("shell", True),),
    )

    with pytest.raises(TimeoutError):
        marker_text = b"marker-text1"
        condition = toolz.flip(bytes.__contains__)(marker_text)
        popened.stdout_peeker.peek_line_until(condition, timeout=1)



import functools
from typing import Callable

import cloudpickle
import numpy as np
import pyarrow as pa
import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    deep_iterable,
    instance_of,
    optional,
)

import xorq.expr.datatypes as dt
import xorq.expr.udf as udf
from xorq.common.utils.name_utils import make_name
from xorq.expr.ml.enums import ResponseMethod
from xorq.expr.ml.structer import KV_ENCODED_TYPE, KVEncoder, Structer
from xorq.expr.udf import make_pandas_expr_udf
from xorq.vendor import ibis


@toolz.curry
def fit_sklearn(df, target=None, *, cls, params):
    obj = cls(**dict(params))
    obj.fit(df, target)
    return obj


@toolz.curry
def fit_sklearn_args(df, *args, cls, params):
    instance = cls(**dict(params))
    instance.fit(df, *args)
    return instance


@toolz.curry
def fit_sklearn_series(df, col, cls, params):
    model = cls(**dict(params))
    model.fit(df[col])
    return model


@toolz.curry
def transform_sklearn(model, df):
    transformed = model.transform(df)
    return transformed


@toolz.curry
def transform_sklearn_series(model, df, col):
    return model.transform(df[col]).toarray().tolist()


@toolz.curry
def transform_sklearn_series_kv(model, df, col):
    return KVEncoder.encode(model, df[col])


@toolz.curry
def transform_sklearn_struct(convert_array, model, df):
    return convert_array(model.transform(df))


@toolz.curry
def predict_sklearn(model, df):
    predicted = model.predict(df)
    return predicted


@toolz.curry
def predict_proba_sklearn(model, df):
    """Predict class probabilities using sklearn model."""
    proba = model.predict_proba(df)
    return list(proba)


@toolz.curry
def decision_function_sklearn(model, df):
    """Compute decision function scores with consistent array output."""
    scores = np.asarray(model.decision_function(df))
    if scores.ndim == 1:
        return [[float(value)] for value in scores]
    return [row.tolist() for row in scores]


@toolz.curry
def feature_importances_sklearn(model, df):
    """Extract feature importances from sklearn model as a single row."""
    importances = model.feature_importances_
    return [importances.tolist()]


@toolz.curry
def kv_encode_output(model, df):
    """
    Convert sklearn transform output to Array[Struct{key, value}] encoded format.

    Enables fully deferred execution by using a fixed return type regardless
    of the actual output schema, which is resolved at execution time via
    get_feature_names_out().
    """
    return KVEncoder.encode(model, df)


@frozen
class DeferredFitOther:
    expr = field(validator=instance_of(ibis.Expr))
    target = field(validator=optional(instance_of(str)))
    features = field(
        validator=optional(deep_iterable(instance_of(str), instance_of(tuple))),
        converter=tuple,
    )
    fit = field(validator=instance_of(Callable))
    other = field(validator=instance_of(Callable))
    return_type = field(validator=instance_of(dt.DataType))
    name_infix = field(validator=instance_of(str))
    cache = field(default=None)

    def __attrs_post_init__(self):
        from xorq.caching import Cache

        if not isinstance(self.cache, (Cache, type(None))):
            raise ValueError(
                f"cache must be of type Optional[Cache], is of type {type(self.cache)}"
            )
        if self.features is None or len(self.features) == 0:
            object.__setattr__(self, "features", tuple(self.expr.schema()))

    def make_name(self, prefix):
        to_tokenize = (self.fit, self.other)
        return make_name(prefix, to_tokenize)

    @property
    def schema(self):
        schema = self.expr.select(self.features).schema()
        return schema

    @property
    def encoded_cols(self):
        return KVEncoder.get_kv_encoded_cols(self.expr, self.features)

    @property
    def fit_schema(self):
        base = self.schema
        if self.target:
            return base | ibis.schema({self.target: self.expr[self.target].type()})
        return base

    @property
    def model_udaf(self):
        return udf.agg.pandas_df(
            fn=toolz.compose(
                cloudpickle.dumps,
                self._inner_fit(
                    fit=self.fit,
                    target=self.target,
                    features=self.features,
                    encoded_cols=self.encoded_cols,
                ),
            ),
            schema=self.fit_schema,
            return_type=dt.binary,
            name=self.make_name(f"fit_{self.name_infix}"),
        )

    @property
    def deferred_model(self):
        deferred_model = self.model_udaf.on_expr(self.expr)
        if self.cache:
            deferred_model = deferred_model.as_table().cache(cache=self.cache)
        return deferred_model

    @functools.cache
    # if we don't cache this, we get extra tags
    def make_deferred_other(self, fn, return_type, name_infix):
        wrapped_fn = self._inner_other(
            other=fn,
            features=self.features,
            return_type=return_type,
            encoded_cols=self.encoded_cols,
        )
        deferred_other = make_pandas_expr_udf(
            computed_kwargs_expr=self.deferred_model,
            fn=wrapped_fn,
            schema=self.schema,
            return_type=return_type,
            name=self.make_name(name_infix),
        )
        return deferred_other

    @property
    def deferred_other(self):
        return self.make_deferred_other(
            fn=self.other,
            return_type=self.return_type,
            name_infix=self.name_infix,
        )

    @property
    def deferred_model_udaf_other(self):
        return (self.deferred_model, self.model_udaf, self.deferred_other)

    @staticmethod
    @toolz.curry
    def _inner_fit(df, fit, target, features, encoded_cols):
        df_decoded, features_decoded = KVEncoder.decode_encoded_columns(
            df, features, encoded_cols
        )
        args = (df_decoded[list(features_decoded)],) + (
            (df_decoded[target],) if target else ()
        )
        return fit(*args)

    @staticmethod
    @toolz.curry
    def _inner_other(model, df, other, features, return_type, encoded_cols):
        df_decoded, features_decoded = KVEncoder.decode_encoded_columns(
            df, features, encoded_cols
        )
        return pa.array(
            other(model, df_decoded[list(features_decoded)]),
            type=return_type.to_pyarrow(),
        )

    @classmethod
    def from_fitted_step(cls, fitted_step, mode=None):
        # Auto-detect mode if not specified (backward compat)
        if mode is None:
            if fitted_step.is_predict and not fitted_step.is_transform:
                mode = ResponseMethod.PREDICT
            else:
                mode = "transform"  # Default to transform

        kwargs = {
            # still need to add: fit, other, return_type, name_infix
            "expr": fitted_step.expr,
            "target": fitted_step.target,
            "features": fitted_step.features,
            "cache": fitted_step.cache,
        }
        if mode == ResponseMethod.PREDICT:
            return cls(
                fit=fit_sklearn(
                    cls=fitted_step.step.typ, params=fitted_step.step.params_tuple
                ),
                other=predict_sklearn,
                return_type=fitted_step.predict_return_type,
                name_infix=ResponseMethod.PREDICT,
                **kwargs,
            )
        elif mode == "transform":
            structer = Structer.from_instance_expr(
                fitted_step.instance, fitted_step.expr, features=fitted_step.features
            )
            target = fitted_step.target if structer.needs_target else None
            match (
                structer.is_series,
                structer.is_kv_encoded,
                structer.struct_has_kv_fields,
            ):
                case (True, True, _):
                    (col,) = fitted_step.features
                    return cls(
                        fit=fit_sklearn_series(
                            col=col,
                            cls=fitted_step.step.typ,
                            params=fitted_step.step.params_tuple,
                        ),
                        other=transform_sklearn_series_kv(col=col),
                        return_type=KV_ENCODED_TYPE,
                        name_infix="transformed_encoded",
                        **kwargs
                        | {
                            "target": target,
                        },
                    )
                case (False, True, _):
                    # Pure KV-encoded output (struct is None)
                    return cls(
                        fit=fit_sklearn_args(
                            cls=fitted_step.step.typ,
                            params=fitted_step.step.params_tuple,
                        ),
                        other=kv_encode_output,
                        return_type=KV_ENCODED_TYPE,
                        name_infix="transformed_encoded",
                        **kwargs,
                    )
                case (False, False, True):
                    # Struct with KV-encoded fields
                    return cls(
                        fit=fit_sklearn_args(
                            cls=fitted_step.step.typ,
                            params=fitted_step.step.params_tuple,
                        ),
                        other=structer.get_convert_struct_with_kv(),
                        return_type=structer.return_type,
                        name_infix="transformed_struct_kv",
                        **kwargs,
                    )
                case _:
                    # Pure struct output (no KV columns)
                    return cls(
                        fit=fit_sklearn_args(
                            cls=fitted_step.step.typ,
                            params=fitted_step.step.params_tuple,
                        ),
                        other=transform_sklearn_struct(structer.get_convert_array()),
                        return_type=structer.return_type,
                        name_infix="transformed",
                        **kwargs,
                    )
        else:
            raise ValueError(f"Unknown mode: {mode}. Must be 'transform' or 'predict'")


@toolz.curry
def deferred_fit_other_sklearn(
    expr,
    target,
    features,
    cls,
    other,
    return_type,
    name_infix,
    params=(),
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=other,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


# Begin specific deferred
@toolz.curry
def deferred_fit_transform(
    expr,
    features,
    fit,
    other,
    return_type,
    target=None,
    name_infix="transform",
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit,
        other=other,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_predict(
    expr,
    target,
    features,
    cls,
    return_type,
    params=(),
    name_infix=ResponseMethod.PREDICT,
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=predict_sklearn,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_transform_sklearn(
    expr,
    target,
    features,
    cls,
    return_type,
    params=(),
    name_infix="transformed",
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=transform_sklearn,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_predict_sklearn(
    expr,
    target,
    features,
    cls,
    return_type,
    params=(),
    name_infix=ResponseMethod.PREDICT,
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=predict_sklearn,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_predict_proba_sklearn(
    expr,
    target,
    features,
    cls,
    return_type,
    params=(),
    name_infix=ResponseMethod.PREDICT_PROBA,
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=predict_proba_sklearn,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_decision_function_sklearn(
    expr,
    target,
    features,
    cls,
    return_type,
    params=(),
    name_infix=ResponseMethod.DECISION_FUNCTION,
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=decision_function_sklearn,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_feature_importances_sklearn(
    expr,
    target,
    features,
    cls,
    return_type,
    params=(),
    name_infix="feature_importances_",
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        target=target,
        features=features,
        fit=fit_sklearn(cls=cls, params=params),
        other=feature_importances_sklearn,
        return_type=return_type,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_transform_series_sklearn(
    expr, col, cls, return_type, params=(), name="transformed", cache=None
):
    return DeferredFitOther(
        expr=expr,
        target=None,
        features=(col,),
        fit=fit_sklearn_series(col=col, cls=cls, params=params),
        other=transform_sklearn_series(col=col),
        return_type=return_type,
        name_infix=name,
        cache=cache,
    )


@toolz.curry
def deferred_fit_transform_series_sklearn_encoded(
    expr, col, cls, params=(), name="transformed_encoded", cache=None
):
    return DeferredFitOther(
        expr=expr,
        target=None,
        features=(col,),
        fit=fit_sklearn_series(col=col, cls=cls, params=params),
        other=transform_sklearn_series_kv(col=col),
        return_type=KV_ENCODED_TYPE,
        name_infix=name,
        cache=cache,
    )


@toolz.curry
def deferred_fit_transform_sklearn_struct(
    expr, features, cls, params=(), target=None, name_infix="transformed", cache=None
):
    structer = Structer.from_instance_expr(cls(**dict(params)), expr, features=features)
    return DeferredFitOther(
        expr=expr,
        features=list(features),
        fit=fit_sklearn_args(cls=cls, params=params),
        other=transform_sklearn_struct(structer.get_convert_array()),
        return_type=structer.return_type,
        target=target,
        name_infix=name_infix,
        cache=cache,
    )


@toolz.curry
def deferred_fit_transform_sklearn_encoded(
    expr,
    features,
    cls,
    params=(),
    target=None,
    name_infix="transformed_encoded",
    cache=None,
):
    return DeferredFitOther(
        expr=expr,
        features=list(features),
        fit=fit_sklearn_args(cls=cls, params=params),
        other=kv_encode_output,
        return_type=KV_ENCODED_TYPE,
        target=target,
        name_infix=name_infix,
        cache=cache,
    )



import datetime
import pathlib
from collections.abc import Mapping, Sequence
from typing import Any, Dict

from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    ParquetTTLSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)
from xorq.vendor.ibis.common.collections import FrozenOrderedDict


def freeze(obj):
    if isinstance(obj, dict):
        return FrozenOrderedDict({k: freeze(v) for k, v in obj.items()})
    elif isinstance(obj, list):
        return tuple(freeze(x) for x in obj)
    elif isinstance(obj, tuple):
        return tuple(freeze(x) for x in obj)
    return obj


class MissingValue:
    def __repr__(self):
        return "<MISSING>"


MISSING = MissingValue()


def deep_diff_objects(obj1, obj2, path="root"):
    differences = []

    if obj1 is not obj2:
        differences.append((path, obj1, obj2))
        return differences

    if isinstance(obj1, Mapping):
        keys1 = set(obj1.keys())
        keys2 = set(obj2.keys())
        for key in keys1 - keys2:
            diff_path = f"{path}.{key}" if path else key
            differences.append((diff_path, obj1[key], MISSING))
        for key in keys2 - keys1:
            diff_path = f"{path}.{key}" if path else key
            differences.append((diff_path, MISSING, obj2[key]))
        for key in keys1 & keys2:
            diff_path = f"{path}.{key}" if path else key
            differences.extend(deep_diff_objects(obj1[key], obj2[key], diff_path))
        return differences

    elif isinstance(obj1, Sequence) and not isinstance(obj1, str):
        if len(obj1) != len(obj2):
            differences.append((path, obj1, obj2))
        for i, (item1, item2) in enumerate(zip(obj1, obj2)):
            diff_path = f"{path}[{i}]"
            differences.extend(deep_diff_objects(item1, item2, diff_path))
        return differences

    else:
        if obj1 != obj2:
            differences.append((path, obj1, obj2))
        return differences


def serialize_ibis_expr(expr):
    try:
        op = expr.op()
    except Exception:
        return repr(expr)

    serialized = {
        "expr_class": expr.__class__.__name__,
        "op_class": op.__class__.__name__,
    }

    op_attrs = {}
    for attr in dir(op):
        if attr.startswith("_"):
            continue
        try:
            value = getattr(op, attr)
        except Exception:
            continue
        if callable(value):
            continue
        op_attrs[attr] = value
    if op_attrs:
        serialized["op_attrs"] = op_attrs

    if hasattr(op, "args"):
        try:
            children = op.args
        except Exception:
            children = None
        if children is not None:
            if isinstance(children, Sequence) and not isinstance(children, str):
                serialized["args"] = [serialize_ibis_expr(child) for child in children]
            else:
                serialized["args"] = serialize_ibis_expr(children)
    return serialized


def diff_ibis_exprs(expr1, expr2):
    if expr1.equals(expr2):
        print("Expressions are equal")
        return

    serialized1 = serialize_ibis_expr(expr1)
    serialized2 = serialize_ibis_expr(expr2)

    diffs = deep_diff_objects(serialized1, serialized2)
    if diffs:
        print("Found differences:")
        for diff in diffs:
            path, val1, val2 = diff
            print(f"At {path}:")
            print(f"  First expression: {val1}")
            print(f"  Second expression: {val2}")
    else:
        print("No differences found (unexpectedly).")

    return diffs


def translate_cache(cache, translation_context: Any) -> Dict:
    if isinstance(cache, SourceCache):
        return {
            "type": "SourceCache",
            "source": cache.storage.source._profile.hash_name,
        }
    elif isinstance(cache, ParquetCache):
        return {
            "type": "ParquetCache",
            "source": cache.storage.source._profile.hash_name,
            "relative_path": str(cache.storage.relative_path),
        }
    elif isinstance(cache, ParquetSnapshotCache):
        return {
            "type": "ParquetSnapshotCache",
            "source": cache.storage.source._profile.hash_name,
            "relative_path": str(cache.storage.relative_path),
        }
    elif isinstance(cache, ParquetTTLSnapshotCache):
        return {
            "type": "ParquetTTLSnapshotCache",
            "source": cache.storage.source._profile.hash_name,
            "relative_path": str(cache.storage.relative_path),
            "ttl_seconds": cache.storage.ttl.total_seconds(),
        }
    elif isinstance(cache, SourceSnapshotCache):
        return {
            "type": "SourceSnapshotCache",
            "source": cache.storage.source._profile.hash_name,
        }
    else:
        raise NotImplementedError(f"Unknown cache type: {type(cache)}")


def load_cache_from_yaml(cache_yaml: Dict, compiler: Any):
    if cache_yaml["type"] == "SourceCache":
        source_profile_name = cache_yaml["source"]
        source = compiler.profiles[source_profile_name]
        return SourceCache.from_kwargs(source=source)
    elif cache_yaml["type"] == "ParquetCache":
        source_profile_name = cache_yaml["source"]
        source = compiler.profiles[source_profile_name]
        return ParquetCache.from_kwargs(
            source=source, relative_path=pathlib.Path(cache_yaml["relative_path"])
        )
    elif cache_yaml["type"] == "ParquetSnapshotCache":
        source_profile_name = cache_yaml["source"]
        source = compiler.profiles[source_profile_name]
        return ParquetSnapshotCache.from_kwargs(
            source=source, relative_path=pathlib.Path(cache_yaml["relative_path"])
        )
    elif cache_yaml["type"] == "ParquetTTLSnapshotCache":
        source_profile_name = cache_yaml["source"]
        source = compiler.profiles[source_profile_name]
        ttl = datetime.timedelta(seconds=cache_yaml["ttl_seconds"])
        return ParquetTTLSnapshotCache.from_kwargs(
            source=source,
            relative_path=pathlib.Path(cache_yaml["relative_path"]),
            ttl=ttl,
        )
    elif cache_yaml["type"] == "SourceSnapshotCache":
        source_profile_name = cache_yaml["source"]
        source = compiler.profiles[source_profile_name]
        return SourceSnapshotCache.from_kwargs(source=source)
    else:
        raise NotImplementedError(f"Unknown cache type: {cache_yaml['type']}")



"""Validation for UDFs.

Warning: This is an experimental module and API here can change without notice.

DO NOT USE DIRECTLY.
"""

from __future__ import annotations

from inspect import Parameter, Signature, signature
from typing import TYPE_CHECKING, Any

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt


if TYPE_CHECKING:
    from collections.abc import Callable


def _parameter_count(funcsig: Signature) -> int:
    """Get the number of positional parameters in a function signature.

    Parameters
    ----------
    funcsig : inspect.Signature
        A UDF signature

    Returns
    -------
    int
        The number of parameters

    """
    kinds = (Parameter.POSITIONAL_OR_KEYWORD, Parameter.POSITIONAL_ONLY)
    return sum(
        param.kind in kinds
        for param in funcsig.parameters.values()
        if param.default is Parameter.empty
    )


def validate_input_type(input_type: list[dt.DataType], func: Callable) -> Signature:
    """Check that the declared number of inputs and signature of func are compatible.

    If the signature of `func` uses *args, then no check is done (since no
    check can be done).
    """
    funcsig = signature(func)
    params = funcsig.parameters.values()

    # We can only do validation if all the positional arguments are explicit
    # (i.e. no *args)
    if not any(param.kind is Parameter.VAR_POSITIONAL for param in params):
        declared_parameter_count = len(input_type)
        function_parameter_count = _parameter_count(funcsig)

        if declared_parameter_count != function_parameter_count:
            raise TypeError(
                f"Function signature {func.__name__!r} has {function_parameter_count:d} parameters, "
                f"input_type has {declared_parameter_count:d}. These must match. Non-column "
                "parameters must be defined as keyword only, i.e., "
                "def foo(col, *, function_param)."
            )

    return funcsig


def validate_output_type(output_type: Any) -> None:
    """Check that the output type is a single datatype."""

    if isinstance(output_type, list):
        raise com.XorqTypeError("The output type of a UDF must be a single datatype.")



from __future__ import annotations

import decimal
import importlib
import importlib.metadata
import operator

import numpy as np
import pandas as pd
import pytest
from packaging.version import parse as vparse

import xorq.vendor.ibis.expr.datatypes as dt
from xorq.backends.pandas import Backend


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        {
            "plain_int64": list(range(1, 4)),
            "plain_strings": list("abc"),
            "plain_float64": [4.0, 5.0, 6.0],
            "plain_datetimes_naive": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=3).values
            ),
            "plain_datetimes_ny": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=3).values
            ).dt.tz_localize("America/New_York"),
            "plain_datetimes_utc": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=3).values
            ).dt.tz_localize("UTC"),
            "plain_uint64": pd.Series(range(1, 4), dtype=np.dtype("uint64")),
            "dup_strings": list("dad"),
            "dup_ints": [1, 2, 1],
            "float64_as_strings": ["100.01", "234.23", "-999.34"],
            "int64_as_strings": list(map(str, range(1, 4))),
            "strings_with_space": [" ", "abab", "ddeeffgg"],
            "translate_from_strings": ["rmz", "abc", "ghj"],
            "translate_to_strings": ["lns", "ovk", "jfr"],
            "int64_with_zeros": [0, 1, 0],
            "float64_with_zeros": [1.0, 0.0, 1.0],
            "float64_positive": [1.0, 2.0, 1.0],
            "strings_with_nulls": ["a", None, "b"],
            "datetime_strings_naive": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=3).values
            ).astype(str),
            "datetime_strings_ny": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=3).values
            )
            .dt.tz_localize("America/New_York")
            .astype(str),
            "datetime_strings_utc": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=3).values
            )
            .dt.tz_localize("UTC")
            .astype(str),
            "decimal": list(map(decimal.Decimal, ["1.0", "2", "3.234"])),
            "array_of_float64": [
                np.array([1.0, 2.0], dtype="float64"),
                np.array([3.0], dtype="float64"),
                np.array([], dtype="float64"),
            ],
            "array_of_int64": [
                np.array([1, 2], dtype="int64"),
                np.array([], dtype="int64"),
                np.array([3], dtype="int64"),
            ],
            "array_of_strings": [
                np.array(["a", "b"], dtype="object"),
                np.array([], dtype="object"),
                np.array(["c"], dtype="object"),
            ],
            "map_of_strings_integers": [{"a": 1, "b": 2}, None, {}],
            "map_of_integers_strings": [{}, None, {1: "a", 2: "b"}],
            "map_of_complex_values": [None, {"a": [1, 2, 3], "b": []}, {}],
        }
    )


@pytest.fixture(scope="module")
def batting_df(data_dir):
    num_rows = 1000
    start_index = 30
    df = pd.read_parquet(data_dir / "parquet" / "batting.parquet").iloc[
        start_index : start_index + num_rows
    ]
    return df.reset_index(drop=True)


@pytest.fixture(scope="module")
def awards_players_df(data_dir):
    return pd.read_parquet(data_dir / "parquet" / "awards_players.parquet")


@pytest.fixture(scope="module")
def df1():
    return pd.DataFrame(
        {"key": list("abcd"), "value": [3, 4, 5, 6], "key2": list("eeff")}
    )


@pytest.fixture(scope="module")
def df2():
    return pd.DataFrame(
        {"key": list("ac"), "other_value": [4.0, 6.0], "key3": list("fe")}
    )


@pytest.fixture(scope="module")
def intersect_df2():
    return pd.DataFrame({"key": list("cd"), "value": [5, 6], "key2": list("ff")})


@pytest.fixture(scope="module")
def time_df1():
    return pd.DataFrame(
        {"time": pd.to_datetime([1, 2, 3, 4]), "value": [1.1, 2.2, 3.3, 4.4]}
    )


@pytest.fixture(scope="module")
def time_df2():
    return pd.DataFrame({"time": pd.to_datetime([2, 4]), "other_value": [1.2, 2.0]})


@pytest.fixture(scope="module")
def time_df3():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=8).values
            ),
            "id": list(range(1, 5)) * 2,
            "value": [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8],
        }
    )


@pytest.fixture(scope="module")
def time_keyed_df1():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(start="2017-01-02 01:02:03.234", periods=6).values
            ),
            "key": [1, 2, 3, 1, 2, 3],
            "value": [1.2, 1.4, 2.0, 4.0, 8.0, 16.0],
        }
    )


@pytest.fixture(scope="module")
def time_keyed_df2():
    return pd.DataFrame(
        {
            "time": pd.Series(
                pd.date_range(
                    start="2017-01-02 01:02:03.234", freq="3D", periods=3
                ).values
            ),
            "key": [1, 2, 3],
            "other_value": [1.1, 1.2, 2.2],
        }
    )


@pytest.fixture(scope="module")
def client(
    df,
    df1,
    df2,
    df3,
    time_df1,
    time_df2,
    time_df3,
    time_keyed_df1,
    time_keyed_df2,
    intersect_df2,
):
    return Backend().connect(
        {
            "df": df,
            "df1": df1,
            "df2": df2,
            "df3": df3,
            "left": df1,
            "right": df2,
            "time_df1": time_df1,
            "time_df2": time_df2,
            "time_df3": time_df3,
            "time_keyed_df1": time_keyed_df1,
            "time_keyed_df2": time_keyed_df2,
            "intersect_df2": intersect_df2,
        }
    )


@pytest.fixture(scope="module")
def df3():
    return pd.DataFrame(
        {
            "key": list("ac"),
            "other_value": [4.0, 6.0],
            "key2": list("ae"),
            "key3": list("fe"),
        }
    )


t_schema = {
    "decimal": dt.Decimal(4, 3),
    "array_of_float64": dt.Array(dt.double),
    "array_of_int64": dt.Array(dt.int64),
    "array_of_strings": dt.Array(dt.string),
    "map_of_strings_integers": dt.Map(dt.string, dt.int64),
    "map_of_integers_strings": dt.Map(dt.int64, dt.string),
    "map_of_complex_values": dt.Map(dt.string, dt.Array(dt.int64)),
}


@pytest.fixture(scope="module")
def t(client):
    return client.table("df", schema=t_schema)


@pytest.fixture(scope="module")
def lahman(batting_df, awards_players_df):
    return Backend().connect(
        {"batting": batting_df, "awards_players": awards_players_df}
    )


@pytest.fixture(scope="module")
def left(client):
    return client.table("left")


@pytest.fixture(scope="module")
def right(client):
    return client.table("right")


@pytest.fixture(scope="module")
def time_left(client):
    return client.table("time_df1")


@pytest.fixture(scope="module")
def time_right(client):
    return client.table("time_df2")


@pytest.fixture(scope="module")
def time_keyed_left(client):
    return client.table("time_keyed_df1")


@pytest.fixture(scope="module")
def time_keyed_right(client):
    return client.table("time_keyed_df2")


@pytest.fixture(scope="module")
def batting(lahman):
    return lahman.table("batting")


@pytest.fixture(scope="module")
def sel_cols(batting):
    cols = batting.columns
    start, end = cols.index("AB"), cols.index("H") + 1
    return ["playerID", "yearID", "teamID", "G"] + cols[start:end]


@pytest.fixture(scope="module")
def players_base(batting, sel_cols):
    return batting[sel_cols].order_by(sel_cols[:3])


@pytest.fixture(scope="module")
def players(players_base):
    return players_base.group_by("playerID")


@pytest.fixture(scope="module")
def players_df(players_base):
    return players_base.execute().reset_index(drop=True)


def compare_versions(module_name, given_version, op):
    try:
        current_version = importlib.metadata.version(module_name)
        return op(vparse(current_version), vparse(given_version))
    except importlib.metadata.PackageNotFoundError:
        return False


def is_older_than(module_name, given_version):
    return compare_versions(module_name, given_version, operator.lt)



import datetime
import functools
import importlib
from contextlib import contextmanager

import toolz
from ibis import Schema as IbisSchema
from ibis.backends import BaseBackend as IbisBaseBackend
from ibis.common.collections import FrozenOrderedDict as IbisFrozenOrderedDict
from ibis.common.temporal import IntervalUnit as IbisIntervalUnit
from ibis.common.temporal import TimestampUnit as IbisTimestampUnit
from ibis.common.temporal import TimeUnit as IbisTimeUnit
from ibis.expr.datatypes import DataType as IbisDataType
from ibis.expr.datatypes.core import Interval as IbisInterval
from ibis.expr.operations.generic import Cast as IbisCast
from ibis.expr.operations.relations import Namespace as IbisNamespace
from ibis.formats.pandas import PandasDataFrameProxy as IbisPandasDataFrameProxy
from ibis.formats.pyarrow import PyArrowTableProxy as IbisPyArrowTableProxy

import xorq.vendor.ibis.expr.operations as ops
from xorq.vendor.ibis import Schema
from xorq.vendor.ibis.backends.profiles import Profile
from xorq.vendor.ibis.common.collections import FrozenOrderedDict
from xorq.vendor.ibis.common.temporal import IntervalUnit, TimestampUnit, TimeUnit
from xorq.vendor.ibis.expr.datatypes import DataType
from xorq.vendor.ibis.expr.datatypes.core import Interval
from xorq.vendor.ibis.expr.operations import Node
from xorq.vendor.ibis.expr.operations.relations import Namespace
from xorq.vendor.ibis.formats.pandas import PandasDataFrameProxy
from xorq.vendor.ibis.formats.pyarrow import PyArrowTableProxy


@functools.singledispatch
def map_ibis(val, kwargs=None):
    try:
        attr = val.__class__.__name__
        module = val.__class__.__module__

        cls = getattr(importlib.import_module(f"xorq.vendor.{module}"), attr)

        kwargs = kwargs or dict(zip(val.argnames, val.args))
        kwargs = toolz.valmap(
            map_ibis,
            kwargs or dict(zip(val.argnames, val.args)),
        )

        return cls(**kwargs)

    except AttributeError as err:
        raise NotImplementedError(f"{type(val)} is not implemented") from err


@map_ibis.register(int)
@map_ibis.register(str)
@map_ibis.register(float)
@map_ibis.register(Node)
@map_ibis.register(type(None))
@map_ibis.register(datetime.datetime)
def map_pass_through(op, kwargs=None):
    return op


@map_ibis.register(tuple)
@map_ibis.register(list)
@map_ibis.register(set)
@map_ibis.register(frozenset)
def map_builtin_container(op, kwargs=None):
    return type(op)(map_ibis(val, None) for val in op)


@map_ibis.register(dict)
def map_builtin_dict(op, kwargs=None):
    return {map_ibis(k, None): map_ibis(v, None) for k, v in op.items()}


@map_ibis.register(IbisCast)
def map_cast(cast, kwargs=None):
    return ops.Cast(arg=map_ibis(cast.arg, None), to=map_ibis(cast.to, None))


@map_ibis.register(IbisSchema)
def map_schema(schema, kwargs=None):
    return Schema(
        dict(zip(schema.names, tuple(map_ibis(typ, kwargs) for typ in schema.types)))
    )


@map_ibis.register(IbisIntervalUnit)
def map_interval_unit(unit, kwargs=None):
    return IntervalUnit(unit.value)


@map_ibis.register(IbisTimestampUnit)
def map_timestamp_unit(unit: IbisTimestampUnit, kwargs=None):
    return TimestampUnit(unit.value)


@map_ibis.register(IbisTimeUnit)
def map_time_unit(unit, kwargs=None):
    return TimeUnit(unit.value)


@map_ibis.register(IbisInterval)
def map_interval(interval, kwargs=None):
    return Interval(
        unit=map_ibis(interval.unit, None), nullable=map_ibis(interval.nullable, None)
    )


@map_ibis.register(IbisDataType)
def map_datatype(datatype, kwargs=None):
    return DataType.from_pyarrow(datatype.to_pyarrow())


@map_ibis.register(IbisPandasDataFrameProxy)
def map_pandas_dataframe_proxy(proxy, kwargs=None):
    return PandasDataFrameProxy(proxy.obj)


@map_ibis.register(IbisPyArrowTableProxy)
def map_pyarrow_table_proxy(proxy, kwargs=None):
    return PyArrowTableProxy(proxy.obj)


@map_ibis.register(IbisFrozenOrderedDict)
def map_frozendict(frozendict, kwargs=None):
    return FrozenOrderedDict(
        tuple(
            (map_ibis(key, kwargs), map_ibis(value, kwargs))
            for key, value in frozendict.items()
        )
    )


backend_registry = {}


@contextmanager
def ibis_utils_backend_registry_context():
    try:
        assert not backend_registry, backend_registry
        yield
    finally:
        backend_registry.clear()


@map_ibis.register(IbisBaseBackend)
def map_backend(backend, kwargs=None):
    backend_id = id(backend)
    if not (new_backend := backend_registry.get(backend_id)):
        backend_registry[backend_id] = new_backend = Profile.from_con(backend).get_con()
        new_backend.con = backend.con
    return new_backend


@map_ibis.register(IbisNamespace)
def map_namespace(namespace, kwargs=None):
    return Namespace(catalog=namespace.catalog, database=namespace.database)


def from_ibis(ibis_expr):
    with ibis_utils_backend_registry_context():
        return ibis_expr.op().replace(map_ibis).to_expr()



import functools
import sys
import tempfile
from pathlib import (
    Path,
)

import pytest

from xorq.common.utils.download_utils import (
    download_xorq_template,
)
from xorq.common.utils.process_utils import (
    Popened,
)
from xorq.ibis_yaml.packager import (
    SdistBuilder,
    Sdister,
    SdistRunner,
)
from xorq.init_templates import InitTemplates


@functools.cache
def get_template_bytes(template=InitTemplates.default):
    with tempfile.TemporaryDirectory() as td:
        target = Path(td).joinpath("template")
        path = download_xorq_template(template, target=target)
        assert target == path
        template_bytes = path.read_bytes()
        return template_bytes


def prep_template_tmpdir(template, tmpdir):
    tmpdir = Path(tmpdir)
    tgz_path = tmpdir.joinpath("template.tar.gz")
    tgz_path.write_bytes(get_template_bytes(template))
    Popened.check_output(f"tar xzvf {tgz_path} --directory {tmpdir}")
    (project_path,) = (el for el in tmpdir.iterdir() if el.name != tgz_path.name)
    return (tgz_path, project_path)


@pytest.mark.slow(level=1)
@pytest.mark.snapshot_check
@pytest.mark.parametrize("template", tuple(InitTemplates))
def test_sdist_path_hexdigest(template, tmpdir, snapshot):
    tgz_path, project_path = prep_template_tmpdir(template, tmpdir)
    sdister = Sdister(project_path)
    actual = sdister.sdist_path_hexdigest
    snapshot.assert_match(actual, f"test_sdist_path_hexdigest-{template}")


@pytest.mark.slow(level=1)
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="requirements.txt issues for python3.10"
)
@pytest.mark.parametrize("template", tuple(InitTemplates))
def test_sdist_builder(template, tmpdir):
    # test that we build and inject the requirements.txt
    tgz_path, project_path = prep_template_tmpdir(template, tmpdir)
    script_path = project_path.joinpath("expr.py")
    sdist_builder = SdistBuilder(script_path=script_path, sdist_path=tgz_path)
    assert sdist_builder.build_path, sdist_builder._uv_tool_run_xorq_build.stderr


@pytest.mark.xfail(reason="depends on release with unique_key optional")
@pytest.mark.slow(level=1)
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="requirements.txt issues for python3.10"
)
@pytest.mark.parametrize("template", tuple(InitTemplates))
def test_sdist_builder_no_requirements(template, tmpdir):
    # test that we build and inject the requirements.txt
    tgz_path, project_path = prep_template_tmpdir(template, tmpdir)
    script_path = project_path.joinpath("expr.py")
    requirements_path = project_path.joinpath("requirements.txt")
    requirements_path.unlink()
    sdist_builder = SdistBuilder.from_script_path(
        script_path=script_path, project_path=project_path, require_requirements=False
    )
    assert sdist_builder.build_path, sdist_builder._uv_tool_run_xorq_build.stderr


@pytest.mark.slow(level=1)
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="requirements.txt issues for python3.10"
)
@pytest.mark.parametrize("template", tuple(InitTemplates))
def test_sdist_builder_no_requirements_fails(template, tmpdir):
    # test that we build and inject the requirements.txt
    tgz_path, project_path = prep_template_tmpdir(template, tmpdir)
    script_path = project_path.joinpath("expr.py")
    requirements_path = project_path.joinpath("requirements.txt")
    requirements_path.unlink()
    #
    sdister = Sdister(project_path=project_path)
    sdist_path = sdister.sdist_path
    bak_path = sdister.sdist_path.with_name(sdist_path.name + ".bak")
    #
    bak_path.rename(sdist_path)
    with pytest.raises(AssertionError):
        sdist_builder = SdistBuilder(
            script_path=script_path, sdist_path=sdist_path, require_requirements=True
        )
        sdist_builder


@pytest.mark.slow(level=2)
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="requirements.txt issues for python3.10"
)
@pytest.mark.parametrize("template", tuple(InitTemplates))
def test_sdist_runner(template, tmpdir):
    tmpdir = Path(tmpdir)
    output_path = tmpdir.joinpath("output")
    tgz_path, project_path = prep_template_tmpdir(template, tmpdir)
    script_path = project_path.joinpath("expr.py")
    sdist_builder = SdistBuilder(script_path=script_path, sdist_path=tgz_path)
    args = (
        "xorq",
        "run",
        "--output-path",
        str(output_path),
        str(sdist_builder.build_path),
    )
    sdist_runner = SdistRunner(sdist_builder.build_path, args=args)
    assert not sdist_runner.popened.popen.wait()
    assert sdist_runner.popened.returncode == 0
    assert output_path.exists()



from pathlib import Path

from xorq.common.utils.caching_utils import get_xorq_cache_dir


def test_default_caching_dir():
    actual_dir = get_xorq_cache_dir()
    assert actual_dir is not None
    assert isinstance(actual_dir, Path)

    assert actual_dir.match("**/.cache/xorq/")



from xorq.common.utils import classproperty


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


default_branch = "main"


class InitTemplates(StrEnum):
    cached_fetcher = "cached-fetcher"
    sklearn = "sklearn"
    penguins = "penguins"

    @classproperty
    def default(self):
        return self.cached_fetcher

    def get_default_branch(template, default_branch=default_branch):
        return dict(templates_branches).get(template, default_branch)


# NOTE: These are commit hashes from when the template update occurred
templates_branches = (
    (InitTemplates.cached_fetcher, "3db774af377cf5b3f0ba1f5b788c7ef7e768042a"),
    (InitTemplates.sklearn, "ee4a2ac1314e8181a49eb3974e2f65516771e0fc"),
    (InitTemplates.penguins, "f6e8417ca5efaeef7f94efb46ed1ef237dfe7862"),
)



from __future__ import annotations

# ruff: noqa: I001

from xorq.vendor.ibis.expr.types.arrays import *  # noqa: F403
from xorq.vendor.ibis.expr.types.binary import *  # noqa: F403
from xorq.vendor.ibis.expr.types.collections import *  # noqa: F403
from xorq.vendor.ibis.expr.types.core import *  # noqa: F403
from xorq.vendor.ibis.expr.types.generic import *  # noqa: F403
from xorq.vendor.ibis.expr.types.geospatial import *  # noqa: F403
from xorq.vendor.ibis.expr.types.inet import *  # noqa: F403
from xorq.vendor.ibis.expr.types.json import *  # noqa: F403
from xorq.vendor.ibis.expr.types.logical import *  # noqa: F403
from xorq.vendor.ibis.expr.types.maps import *  # noqa: F403
from xorq.vendor.ibis.expr.types.numeric import *  # noqa: F403
from xorq.vendor.ibis.expr.types.relations import *  # noqa: F403
from xorq.vendor.ibis.expr.types.joins import *  # noqa: F403
from xorq.vendor.ibis.expr.types.strings import *  # noqa: F403
from xorq.vendor.ibis.expr.types.structs import *  # noqa: F403
from xorq.vendor.ibis.expr.types.temporal import *  # noqa: F403
from xorq.vendor.ibis.expr.types.uuid import *  # noqa: F403



import re
import shutil
import sys
from itertools import chain
from pathlib import Path

import pandas as pd
import pytest

import xorq.api as xo
from xorq.cli import (
    OutputFormats,
    build_command,
    run_command,
)
from xorq.common.utils.node_utils import (
    find_node,
)
from xorq.common.utils.process_utils import (
    Popened,
    remove_ansi_escape,
    subprocess_run,
)
from xorq.flight.client import (
    FlightClient,
)
from xorq.ibis_yaml.compiler import (
    build_expr,
    load_expr,
)
from xorq.init_templates import InitTemplates


build_run_examples_expr_names = (
    ("local_cache.py", "expr"),
    ("multi_engine.py", "expr"),
    ("remote_caching.py", "expr"),
    ("iris_example.py", "expr"),
    ("simple_example.py", "expr"),
    ("deferred_read_csv.py", "pg_expr_replace"),
    ("train_test_splits.py", "train_table"),
    ("train_test_splits.py", "split_column"),
    ("postgres_caching.py", "expr"),
    ("xgboost_udaf.py", "expr"),
    ("expr_scalar_udf.py", "expr"),
    ("bank_marketing.py", "encoded_test"),
    ("flight_udtf_llm_example.py", "expr"),
    ("pyiceberg_backend_simple.py", "expr"),
    ("python_udwf.py", "expr"),
)


def test_build_command_function(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.py"

    build_command(script_path, "expr", str(builds_dir))
    assert builds_dir.exists()


def test_build_command(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_build_command_with_udtf(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "udxf_expr.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "-e",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)
    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_build_command_on_notebook(monkeypatch, tmp_path, fixture_dir, capsys):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.ipynb"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "-e",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_build_command_with_cache_dir(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    cache_dir = tmp_path / "cache"
    script_path = fixture_dir / "pipeline.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(builds_dir),
        "--cache-dir",
        str(cache_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_run_command_default(tmp_path, fixture_dir):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()
        test_args = [
            "xorq",
            "run",
            expression_path,
        ]
        (returncode, _, _) = subprocess_run(test_args)

        # test with problematic name (see https://github.com/xorq-labs/xorq/issues/1116)
        test_args = [
            "xorq",
            "run",
            str(
                shutil.move(
                    expression_path,
                    Path(expression_path).parent.joinpath("becb4e71406b.bak"),
                )
            ),
        ]
        (returncode, _, stderr) = subprocess_run(test_args)

        assert returncode == 0, stderr
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize("output_format", OutputFormats)
def test_run_command(tmp_path, fixture_dir, output_format):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / f"test.{output_format}"
        expression_path = match.group()
        test_args = [
            "xorq",
            "run",
            expression_path,
            "--output-path",
            str(output_path),
            "--format",
            output_format,
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0

    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize("output_format", ["csv", "json", "parquet"])
def test_run_cached_command(tmp_path, fixture_dir, output_format):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / f"test.{output_format}"
        expression_path = match.group()
        test_args = [
            "xorq",
            "run-cached",
            expression_path,
            "--output-path",
            str(output_path),
            "--format",
            output_format,
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
def test_run_cached_command_default(tmp_path, fixture_dir):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()
        test_args = [
            "xorq",
            "run-cached",
            expression_path,
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize(
    "cache_type,ttl",
    [
        ("modification-time", None),
        ("snapshot", None),
        ("snapshot", "3600"),
    ],
)
def test_run_cached_command_cache_types(tmp_path, fixture_dir, cache_type, ttl):
    target_dir = tmp_path / "build"
    cache_dir = tmp_path / "cache"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / "test.parquet"
        expression_path = match.group()
        test_args = [
            "xorq",
            "run-cached",
            expression_path,
            "--cache-dir",
            str(cache_dir),
            "--cache-type",
            cache_type,
            "--output-path",
            str(output_path),
        ]
        if ttl is not None:
            test_args.extend(["--ttl", ttl])
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.xdist_group(name="serve")
@pytest.mark.parametrize(
    "host,port,cache_dir", [(None, None, None), ("localhost", None, "cache")]
)
def test_serve_command(tmp_path, fixture_dir, cache_dir, host, port):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "udxf_pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (return_code, stdout, stderr) = subprocess_run(build_args)
    assert return_code == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()

        optional_args = tuple(
            chain.from_iterable(
                (arg, value)
                for arg, value in (
                    ("--cache-dir", str(tmp_path / cache_dir) if cache_dir else None),
                    ("--host", host),
                    ("--port", port),
                )
                if value
            )
        )

        serve_args = ("xorq", "serve-flight-udxf", str(expression_path), *optional_args)

        serve_process = Popened(serve_args)
        port = peek_port(serve_process)

        flight_con = xo.flight.connect(host=host, port=int(port))
        assert (
            serve_process.popen.poll() is None
            and "diamonds_exchange_command" in flight_con.list_exchanges()
        )
        serve_process.popen.terminate()

    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize("output_format", ["csv", "json", "parquet"])
def test_run_command_stdout(tmp_path, fixture_dir, output_format):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"
    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()
        test_args = [
            "xorq",
            "run",
            expression_path,
            "--output-path",
            "-",
            "--format",
            output_format,
        ]
        (returncode, stdout, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert stdout
    else:
        raise AssertionError("No expression hash")


def test_run_command_logging(tmp_path):
    """Test that run_command emits expected structured log events with metrics."""
    from unittest.mock import MagicMock, patch

    import pyarrow.parquet as pq

    # Build a self-contained memtable expression (no postgres needed)
    expr = xo.memtable({"a": [1, 2, 3], "b": [4, 5, 6]}, name="test_table")
    expr_path = build_expr(
        expr, builds_dir=tmp_path / "builds", cache_dir=tmp_path / "cache"
    )
    output_path = tmp_path / "out.parquet"

    mock_logger = MagicMock()
    with patch("xorq.common.utils.logging_utils.get_logger", return_value=mock_logger):
        run_command(str(expr_path), str(output_path), "parquet")

    info_calls = mock_logger.info.call_args_list
    event_names = [c.args[0] for c in info_calls]
    assert "run.start" in event_names
    assert "run.expr_loaded" in event_names
    assert "run.done" in event_names

    # run.start captures all input params
    start_call = next(c for c in info_calls if c.args[0] == "run.start")
    assert start_call.kwargs["expr_path"] == str(expr_path)
    assert start_call.kwargs["output_path"] == str(output_path)
    assert "output_format" in start_call.kwargs
    assert "limit" in start_call.kwargs

    # run.expr_loaded has elapsed timing
    loaded_call = next(c for c in info_calls if c.args[0] == "run.expr_loaded")
    assert "elapsed_s" in loaded_call.kwargs
    assert loaded_call.kwargs["elapsed_s"] >= 0

    # run.done carries timing + output_format (from timed)
    done_call = next(c for c in info_calls if c.args[0] == "run.done")
    assert "elapsed_s" in done_call.kwargs
    assert "output_format" in done_call.kwargs

    # run.output_written carries file metrics for parquet output
    written_call = next(c for c in info_calls if c.args[0] == "run.output_written")
    assert "bytes" in written_call.kwargs
    assert written_call.kwargs["bytes"] > 0
    assert "rows" in written_call.kwargs
    assert written_call.kwargs["rows"] == 3
    # rows uses pq.read_metadata (parquet footer) — verify it matches the actual file
    assert pq.read_metadata(output_path).num_rows == written_call.kwargs["rows"]


def test_run_command_error_logging(tmp_path):
    """Test that run_command logs run.failed and re-raises exceptions."""
    from unittest.mock import MagicMock, patch

    mock_logger = MagicMock()
    nonexistent_path = tmp_path / "does_not_exist"
    with patch("xorq.common.utils.logging_utils.get_logger", return_value=mock_logger):
        with pytest.raises(Exception):  # noqa: B017
            run_command(str(nonexistent_path), str(tmp_path / "out.parquet"))

    assert mock_logger.exception.called
    exception_call = mock_logger.exception.call_args
    assert exception_call.args[0] == "run.failed"
    assert "error" in exception_call.kwargs


@pytest.mark.parametrize(
    "expression,message",
    [
        ("integer", "The object integer must be an instance of"),
        ("missing", "Expression missing not found"),
    ],
)
@pytest.mark.slow(level=1)
def test_build_command_bad_expr_name(tmp_path, fixture_dir, expression, message):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "-e",
        expression,
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)
    assert returncode != 0
    assert message in stderr.decode("ascii")


@pytest.mark.parametrize(
    ("example", "expr_name"),
    build_run_examples_expr_names,
)
@pytest.mark.slow(level=2)
def test_examples(
    example,
    expr_name,
    examples_dir,
    tmp_path,
):
    # build
    builds_dir = tmp_path / "builds"
    example_path = examples_dir / example
    assert example_path.exists()
    build_args = (
        "xorq",
        "build",
        str(example_path),
        "--expr-name",
        expr_name,
        "--builds-dir",
        str(builds_dir),
    )
    print(" ".join(build_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr
    print(stderr.decode("ascii"), file=sys.stderr)
    expression_path = Path(stdout.decode("ascii").strip().split("\n")[-1])
    # debugging can capture stdout and result in spurious path of "."
    assert expression_path.name and expression_path.exists()

    # run
    output_format = "parquet"
    output_path = expression_path / f"test.{output_format}"
    assert not output_path.exists()
    run_args = (
        "xorq",
        "run",
        str(expression_path),
        "--format",
        output_format,
        "--output-path",
        str(output_path),
    )
    print(" ".join(run_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(run_args)
    assert returncode == 0, stderr
    print(stderr, file=sys.stderr)
    assert output_path.exists()


def test_init_command_default(tmpdir):
    path = Path(tmpdir).joinpath("xorq-template-default")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
    )
    print(" ".join(init_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()


@pytest.mark.parametrize("template", InitTemplates)
def test_init_command_sklearn(template, tmpdir):
    path = Path(tmpdir).joinpath(f"xorq-template-{template}")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        template,
    )
    print(" ".join(init_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()


@pytest.mark.parametrize("template", InitTemplates)
def test_init_command_path_exists(template, tmpdir):
    path = Path(tmpdir).joinpath(f"xorq-template-{template}")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        template,
    )
    print(" ".join(init_args), file=sys.stderr)
    path.mkdir()
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode != 0


@pytest.mark.slow(level=2)
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="requirements.txt issues for python3.10"
)
@pytest.mark.parametrize("template", InitTemplates)
def test_init_uv_build_uv_run(template, tmpdir):
    tmpdir = Path(tmpdir)
    path = tmpdir.joinpath(f"xorq-template-{template}")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        template,
    )
    print(" ".join(init_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()
    assert path.joinpath("requirements.txt").exists()

    build_args = (
        "xorq",
        "uv-build",
        str(path.joinpath("expr.py")),
    )
    (returncode, stdout, stderr) = subprocess_run(build_args, do_decode=True)
    assert returncode == 0, stderr
    build_path = Path(stdout.strip().split("\n")[-1])
    assert build_path.exists()

    output_path = tmpdir.joinpath("output")
    run_args = (
        "xorq",
        "uv-run",
        "--output-path",
        str(output_path),
        str(build_path),
    )
    (returncode, stdout, stderr) = subprocess_run(run_args, do_decode=True)
    assert returncode == 0, stderr
    assert output_path.exists()


serve_hashes = (
    "1e5f8b2b262dd841f1db1bc19ac6abc2",  # batting, rel.Read
    "e210634ae2df82a56dd49d7cf7516d14",  # awards_players, rel.Read
    "edaa2a90ba8df1688e67a24edb03964e",  # left, ops.Filter
    "929550bdf9130ece976a83ae2cf285f8",  # right, ops.DropColumns
)


@pytest.fixture(scope="session")
def pipeline_https_build(tmp_path_factory, fixture_dir):
    builds_dir = tmp_path_factory.mktemp("builds")
    script_path = fixture_dir / "pipeline_https.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args, do_decode=True)

    assert "Building expr" in stderr
    assert returncode == 0, stderr
    assert builds_dir.exists()
    serve_dir = Path(stdout.strip())
    return serve_dir


def peek_port(popened, timeout=60):
    def do_match(buf):
        (*_, line) = remove_ansi_escape(buf.decode("ascii").strip()).rsplit("\n", 1)
        match = re.match(".*on grpc://localhost:(\\d+)$", line)
        return match

    popened.popen.poll()
    if popened.popen.returncode:
        raise Exception(popened.stderr)
    try:
        buf = popened.stdout_peeker.peek_line_until(do_match, timeout=timeout)
    except TimeoutError as e:
        popened.popen.terminate()
        raise Exception(popened.stderr) from e
    (as_string,) = do_match(buf).groups()
    port = int(as_string)
    return port


def hit_server(port, expr):
    client = FlightClient(port=port)
    (_, rbr) = client.do_exchange("default", expr)
    df = rbr.read_pandas()
    return df


@pytest.mark.slow(level=3)
@pytest.mark.parametrize("serve_hash", serve_hashes)
def test_serve_unbound_hash(serve_hash, pipeline_https_build):
    lookup = {
        "929550bdf9130ece976a83ae2cf285f8": "xorq.vendor.ibis.expr.operations.DropColumns",
        "edaa2a90ba8df1688e67a24edb03964e": "xorq.vendor.ibis.expr.operations.Filter",
    }
    expr = load_expr(pipeline_https_build)
    typ = lookup.get(serve_hash)
    subexpr = find_node(expr, hash=serve_hash, tag=None, typs=typ).to_expr()

    serve_args = (
        "xorq",
        "serve-unbound",
        str(pipeline_https_build),
        "--to_unbind_hash",
        serve_hash,
    ) + (("--typ", typ) if typ else ())
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)
    actual = hit_server(port=port, expr=subexpr)
    expected = expr.execute()
    (actual, expected) = (
        df.sort_values(list(df.columns), ignore_index=True) for df in (actual, expected)
    )
    assert actual.equals(expected)

    serve_popened.popen.terminate()


serve_tags = (
    "read-batting",
    "read-players",
    "batting-filtered",
    "players-filtered",
    # this needs the fix for finding the correct source
    "joined",
)


@pytest.mark.slow(level=3)
@pytest.mark.parametrize("serve_tag", serve_tags)
def test_serve_unbound_tag(serve_tag, pipeline_https_build):
    expr = load_expr(pipeline_https_build)
    subexpr = find_node(expr, hash=None, tag=serve_tag).to_expr()

    serve_args = (
        "xorq",
        "serve-unbound",
        str(pipeline_https_build),
        "--to_unbind_tag",
        serve_tag,
    )
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)
    actual = hit_server(port=port, expr=subexpr)
    expected = expr.execute()
    (actual, expected) = (
        df.sort_values(list(df.columns), ignore_index=True) for df in (actual, expected)
    )
    assert actual.equals(expected)

    serve_popened.popen.terminate()


@pytest.mark.slow(level=1)
def test_serve_unbound_tag_get_exchange(pipeline_https_build, parquet_dir):
    batting_url = "https://storage.googleapis.com/letsql-pins/batting/20240711T171118Z-431ef/batting.parquet"
    serve_tag = "read-batting"
    expr = load_expr(pipeline_https_build)

    serve_args = (
        "xorq",
        "serve-unbound",
        str(pipeline_https_build),
        "--to_unbind_tag",
        serve_tag,
    )
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)

    flight_backend = xo.flight.connect(port=port)
    f = flight_backend.get_exchange("default")
    actual = xo.deferred_read_parquet(batting_url).pipe(f).execute()

    expected = expr.execute()
    (actual, expected) = (
        df.sort_values(list(df.columns), ignore_index=True) for df in (actual, expected)
    )
    assert actual.equals(expected)

    serve_popened.popen.terminate()


@pytest.mark.slow(level=1)
def test_serve_unbound_tag_get_exchange_udf(fixture_dir, tmp_path):
    import pandas as pd

    df = pd.DataFrame([float(v) for v in range(10)], columns=["x"])

    serve_tag = "full"

    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline_pandas_udf.py"

    import contextlib
    import io

    # Capture print output
    output = io.StringIO()

    with contextlib.redirect_stdout(output):
        build_command(script_path, "expr", str(builds_dir))

    serve_args = (
        "xorq",
        "serve-unbound",
        str(output.getvalue().strip()),
        "--to_unbind_tag",
        serve_tag,
    )
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)

    flight_backend = xo.flight.connect(port=port)
    f = flight_backend.get_exchange("default")
    actual = xo.connect().register(df).select("x").pipe(f).execute()

    assert not actual.empty

    serve_popened.popen.terminate()


@pytest.mark.slow(level=3)
def test_serve_penguins_template(tmpdir, tmp_path):
    tmpdir = Path(tmpdir)
    path = tmpdir.joinpath("xorq-template-penguins")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        "penguins",
    )

    (returncode, stdout, stderr) = subprocess_run(init_args)

    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()
    assert path.joinpath("requirements.txt").exists()

    target_dir = tmp_path / "build"
    build_args = [
        "xorq",
        "build",
        str(path / "expr.py"),
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        serve_hash = "e4122ce163e8680c757d6a5ed665440a"  # RemoteTable (test split)

        serve_args = (
            "xorq",
            "serve-unbound",
            str(target_dir / match.group()),
            "--to_unbind_hash",
            serve_hash,
        )
        serve_popened = Popened(serve_args, deferred=False)
        port = peek_port(serve_popened)

        # Create sample penguin data using memtable instead of reading from URL
        sample_data = pd.DataFrame(
            {
                "bill_length_mm": [
                    39.1,
                    39.5,
                    40.3,
                    36.7,
                    39.3,
                    38.9,
                    39.2,
                    34.1,
                    42.0,
                    37.8,
                ],
                "bill_depth_mm": [
                    18.7,
                    17.4,
                    18.0,
                    19.3,
                    20.6,
                    17.8,
                    19.6,
                    18.1,
                    20.2,
                    17.1,
                ],
                "species": [
                    "Adelie",
                    "Adelie",
                    "Adelie",
                    "Adelie",
                    "Adelie",
                    "Chinstrap",
                    "Chinstrap",
                    "Chinstrap",
                    "Gentoo",
                    "Gentoo",
                ],
            }
        )

        expr = xo.memtable(sample_data, name="penguins")

        actual = hit_server(port=port, expr=expr)
        assert not actual.empty
        assert actual["predict"].isin(("Adelie", "Chinstrap", "Gentoo")).all()
        assert len(actual) == len(sample_data)
    else:
        raise AssertionError("No expression hash")
